package leds;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

/**
 * This class maps memory to the OLED display interface provided by the asus_oled kernel driver.
 * Check the <a target=_blank href="http://lapsus.berlios.de/asus_oled.html">oficial asus_oled page</a> for details.
 * 
 * @author costing
 * @since Oct 7, 2008
 */
public class Canvas {

	/**
	 * Raw pixel bitmap
	 */
	private final char area[][];

	/**
	 * Cached write mode
	 */
	private WriteMode writeMode = null;
	
	/**
	 * Model type. Possible values are the {@link Model} enum values.
	 */
	private final Model mModel;
	
	/**
	 * The width of the display, in pixels.
	 */
	public final int W;
	
	/**
	 * The height of the display, in pixels.
	 */
	public final int H;
	
	/**
	 * Width, in chars
	 */
	private final int X;
	/**
	 * Height, in chars
	 */
	private final int Y;
	
	/**
	 * Constant for aligning text to the left.
	 * @see #drawString(Object, int, int, int)
	 * @see #drawString(String, int, int, int) 
	 */
	public static final int ALIGN_LEFT = 0;
	
	/**
	 * Constant for aligning text to the right.
	 * @see #drawString(Object, int, int, int)
	 * @see #drawString(String, int, int, int) 
	 */	
	public static final int ALIGN_RIGHT = 1;
	
	/**
	 * Constant for centering text.
	 * @see #drawString(Object, int, int, int)
	 * @see #drawString(String, int, int, int) 
	 */
	public static final int ALIGN_CENTER = 2;
	
	/**
	 * Image mapped by default on the device
	 */
	private final BufferedImage image;
	
	/**
	 * The Graphics2D instance of the image
	 */
	private final Graphics2D graphics;
	
	/**
	 * Create a default Canvas object for the OLED device returned by {@link Utils#getModel()} 
	 */
	public Canvas(){
		this(Utils.getModel());
	}
	
	/**
	 * Allocate memory for the image.
	 * @param model OLED model
	 */
	public Canvas(final Model model){
		this.mModel = model;
		
		this.W = this.mModel.getWidth();
		this.H = this.mModel.getHeight();
				
		this.X=this.W/8;
		
		this.Y=this.H;
		
		this.area = new char[this.X][this.Y];
		
		clear();
		
		this.image = new BufferedImage(this.W, this.H, BufferedImage.TYPE_INT_RGB);
		
		this.graphics = (Graphics2D) this.image.getGraphics();
		
		clearImage();
	}
	
	
	/**
	 * Get the BufferedImage that will be dumped on the display
	 * 
	 * @return the image
	 * @see #writeImage()
	 */
	public BufferedImage getImage(){
		return this.image;
	}
	
	/**
	 * Get the Graphics2D object that can be used to draw on the display.
	 * 
	 * @return the Graphics2D
	 * @see #writeImage()
	 */
	public Graphics2D getGraphics(){
		return this.graphics;
	}
	
	/**
	 * Reset the internal Image to all black and writing in white.
	 */
	public void clearImage(){
		this.graphics.setColor(Color.BLACK);
		this.graphics.fillRect(0, 0, this.W, this.H);
		this.graphics.setColor(Color.WHITE);
	}
	
	/**
	 * Make the screen black
	 */
	public void clear(){
		fill(false);
	}
	
	/**
	 * Fill the entire canvas.
	 * 
	 * @param bFill true to lit all the leds, false to turn them off
	 */
	public void fill(final boolean bFill){
		for (int x=0; x<this.X; x++)
			for (int y=0; y<this.Y; y++)
				this.area[x][y] = bFill ? 0 : (char)255;
	}

	/**
	 * Create the negative of the current image in memory.
	 */
	public void invert(){
		for (int x=this.X-1; x>=0; x--)
		    for (int y=this.Y-1; y>=0; y--)
			this.area[x][y] = (char) (255 - this.area[x][y]); 
	}
	
	/**
	 * Fill an area of LEDs with either on or off 
	 * 
	 * @param x1 start of area (x1&lt;=x2)
	 * @param y1 start of area (y1&lt;=y2)
	 * @param x2 end of area (x1&lt;=x2)
	 * @param y2 end of area (y1&lt;=y2)
	 * @param bFill true for on, false for off
	 */
	public void fill(final int x1, final int y1, final int x2, final int y2, final boolean bFill){
		for (int x=x1; x<x2; x++)
			for (int y=y1; y<y2; y++)
				set(x, y, bFill);
	}
	
	/**
	 * Set an individual led to on or off
	 * 
	 * @param x X coordinate
	 * @param y Y coordinate
	 * @param bState true for on, false for off
	 */
	public void set(final int x, final int y, final boolean bState){
		final int xReal = this.X*8 - x - 1;
		
		if (bState){
			// set the bit to 0 
			final byte mask = (byte) (0xFF ^ (1<<(7-xReal%8)));
			this.area[xReal/8][y] &= mask;
		}
		else
			this.area[xReal/8][y] |= 1<<(7-xReal%8);
	}
	
//	private static final String vsBits[] = {"0000", "0001", "0010", "0011", "0100", "0101", "0110", "0111",
//						"1000", "1001", "1010", "1011", "1100", "1101", "1110", "1111"};

	/**
	 * 4 bits, in reversed order (how they are mapped internally on the device)
	 */
	private static final String vsRevBits[] = {
		"0000", //$NON-NLS-1$
		"1000", //$NON-NLS-1$
		"0100", //$NON-NLS-1$
		"1100", //$NON-NLS-1$
		"0010", //$NON-NLS-1$
		"1010", //$NON-NLS-1$
		"0110", //$NON-NLS-1$
		"1110", //$NON-NLS-1$
		"0001", //$NON-NLS-1$
		"1001", //$NON-NLS-1$
		"0101", //$NON-NLS-1$
		"1101", //$NON-NLS-1$
		"0011", //$NON-NLS-1$
		"1011", //$NON-NLS-1$
		"0111", //$NON-NLS-1$
		"1111"  //$NON-NLS-1$
	};

//	private static final String toBits(final char c){
//		return vsBits[c>>4]+vsBits[c&0x0f];
//	}
	
	/**
	 * Convert a byte to the string in bits, so that the human readable output matches what you see on the device in the end
	 * 
	 * @param c character to convert
	 * @return String representation of the byte
	 */
	private static final String toRevBits(final char c){
		return vsRevBits[c&0x0f]+vsRevBits[c>>4];
	}
	
	/** 
	 * Get the legacy asus_oled format.
	 */
	@Override
	public String toString(){
		final StringBuilder sb = new StringBuilder( (this.X+1)*this.Y + 20 );

		sb.append("<s:").append(this.X*8).append('x').append(this.Y).append(">\n");  //$NON-NLS-1$//$NON-NLS-2$

		for (int y=0; y<this.Y; y++){
			for (int x=this.X-1; x>=0; x--){
				sb.append(toRevBits((char)(255-this.area[x][y])));
			}
			
			sb.append('\n');
		}
		
		return sb.toString();
	}
	
	/** 
	 * Get the enhanced asus_oled format. This format will be used by default to talk to the driver.
	 * @return the binary mapping of the bits, in asus_oled format
	 * @see #write() 
	 */
	public byte[] toBinaryArray(){
		byte[] ret = new byte[this.X*this.Y+1];

		ret[0] = 'b';		
		
		for (int y=0; y<this.Y; y++)
			for (int x=0; x<this.X; x++)
				ret[y*this.X + x + 1 ] = (byte) this.area[x][y];
		
		return ret;
	}
	
	/** 
	 * Write the contents of the image to the device.
	 *  
	 * @return true if the writing was OK, false if there was an error (driver not loaded?) 
	 */
	public boolean write(){
		final WriteMode wm = getWriteMode();
		
		if (wm==WriteMode.ASUS_OLED_004)
			return write(toBinaryArray());
		else
		if (wm==WriteMode.ASUS_OLED_003)
			return write(toString().getBytes());
		
		return false;
	}
	
	/**
	 * Commit the internal bufferedImage to the display.
	 * 
	 * @return true if writing was possible, false if not (asus_oled driver not loaded?)
	 * @see #getImage()
	 */
	public boolean writeImage(){
		copyImage(this.image);

		return write();
	}
	
	/**
	 * Synchronized writing to the device to avoid getting weird results
	 * 
	 * @param data data to send to the device
	 * @return true if writing was successful
	 */
	private static synchronized boolean write(final byte[] data){
		final File device;
		
		try{
			device = getDevice();
		}
		catch (Exception e){
			System.err.println("Exception determining the device: "+e); //$NON-NLS-1$
			e.printStackTrace();
			return false;
		}
		
		if (device == null)
			return false;
		
		FileOutputStream fos = null;
		
		try{
			fos = new FileOutputStream(device);

			fos.write(data, 0, data.length);
			fos.flush();

			return true;
		}
		catch (IOException ioe){
			return false;
		}
		finally{
			try{
				if (fos!=null)
					fos.close();
			}
			catch (IOException ioe){
				// ignore
			}
		}
	}
	
	/**
	 * Last used device
	 */
	private static File fDevice = null;
	
	/**
	 * Get the virtual file to write the contents to
	 * 
	 * @return the file, or <code>null</code>
	 */
	private static File getDevice(){
		if (fDevice==null || !fDevice.exists() || !fDevice.isFile() || !fDevice.canWrite()){
			if (fDevice!=null){
				System.err.println("Device '"+fDevice.getAbsolutePath()+"' is no longer available, looking for options");  //$NON-NLS-1$//$NON-NLS-2$
			}
			
			fDevice = null;
			
			final File fDir = new File("/sys/class/asus_oled"); //$NON-NLS-1$
			
			String[] list = fDir.list();
			
			for (String s: list){
				if (s.startsWith("oled_")){ //$NON-NLS-1$
					final String sPath = "/sys/class/asus_oled/"+s+"/picture";   //$NON-NLS-1$//$NON-NLS-2$
					
					fDevice = new File(sPath);
					
					if (fDevice.exists() && fDevice.isFile() && fDevice.canWrite()){
						System.err.println("Found an ASUS OLED device under '"+sPath+'\''); //$NON-NLS-1$
						
						return fDevice;
					}
	
					System.err.println("File "+sPath+" is not writable");  //$NON-NLS-1$//$NON-NLS-2$
				
					fDevice = null;
				}
			}
		}
		
		return fDevice;
	}
	
	/**
	 * Copy the contents of a BufferedImage. It will take only at most 256x32 pixels 
	 * from the original image, starting from 0x0. Any pixel that is not completely
	 * black will be turned on on the OLED display. 
	 * 
	 * @param bi the image to display
	 */
	public void copyImage(final BufferedImage bi){
		if (bi==null)
		    return;
		
		for (int y=0; y<this.image.getHeight() && y<this.Y; y++)
		    for (int x=0; x<this.image.getWidth() && x<this.X*8; x++){
		    	int rgb = this.image.getRGB(x, y) & 0xFFFFFF;
			
		    	set(x, y, rgb > 0);
		    }
	}
	
	/**
	 * Copy as much as possible from the given image, applying it from (x,y).
	 * 
	 * @param bi image to copy to the internal BufferedImage object
	 * @param x starting x coordinate
	 * @param y starting y coordinate
	 */
	public void apply(final BufferedImage bi, final int x, final int y){
		this.image.getGraphics().drawImage(bi, x, y, null);
	}
	
	/**
	 * Print a string starting from (x,y).
	 * 
	 * @param s String to display
	 * @param x X starting point
	 * @param y Y starting point (the base of the characters will be here)
	 */
	public void drawString(final String s, final int x, final int y) {
		drawString(s, x, y, ALIGN_LEFT);
	}

	/**
	 * Print a string with a given alignment.
	 * 
	 * @param s string to display
	 * @param x X starting point
	 * @param y Y starting point (the base of the characters will be here)
	 * @param align one of {@link #ALIGN_LEFT}, {@link #ALIGN_CENTER} or {@link #ALIGN_RIGHT}
	 */
	public void drawString(final String s, final int x, final int y, final int align) {
		int xStart = x;

		if (align != ALIGN_LEFT) {
			int advance = getTextWidth(s);

			xStart -= advance / (align == ALIGN_CENTER ? 2 : 1);
		}

		this.graphics.drawString(s, xStart, y);
	}

	/**
	 * For the current font set in the Graphics component of the internal image, get
	 * the width, in pixels, of a given text.
	 * 
	 * @param text string that will be displayed
	 * @return width of the text, in pixels
	 */
	public int getTextWidth(final String text){
		final FontMetrics metrics = this.graphics.getFontMetrics();
		
		return metrics.charsWidth(text.toCharArray(), 0, text.length());
	}
	
	/**
	 * For the current font set in the Graphics component of the internal image, get
	 * the height of any text, in pixels. 
	 * 
	 * @return height of the letters
	 * @see FontMetrics#getAscent()
	 */
	public int getTextHeight(){
		final FontMetrics metrics = this.graphics.getFontMetrics();
		
		return metrics.getAscent();
	}
	
	/**
	 * Print the toString() of any object from (x,y).
	 * 
	 * @param o object to be displayed as text
	 * @param x X starting point
	 * @param y Y starting point (the base of the characters will be here)
	 */
	public void drawString(final Object o, final int x, final int y) {
		if (o!=null)
			drawString(o.toString(), x, y, ALIGN_LEFT);
	}

	/**
	 * Print the toString() of any object from (x,y) with the given alignment. 
	 * 
	 * @param o object to be displayed as text
	 * @param x X starting point
	 * @param y Y starting point (the base of the characters will be here)
	 * @param align one of {@link #ALIGN_LEFT}, {@link #ALIGN_CENTER} or {@link #ALIGN_RIGHT}
	 * @see #drawString(String, int, int, int)
	 */
	public void drawString(final Object o, final int x, final int y, final int align) {
		if (o!=null)
			drawString(o.toString(), x, y, align);
	}
	
	/**
	 * Get the first numerical value from a version string
	 * 
	 * @param sVersion version string
	 * @return major.minor
	 */
	private static final float getVersion(final String sVersion){
		float ver = 0;
		
		if (sVersion==null || sVersion.length()==0)
			return ver;
		
		char c;
		int idx = 0;
		
		while ( idx<sVersion.length() && ((c=sVersion.charAt(idx)) < '0' || c>'9' ) )
			idx ++;
		
		float major = 0;
		float minor = 0;
		float factor = 1;
		
		while ( idx<sVersion.length() && (c=sVersion.charAt(idx))>='0' && c<='9' ){
			major = major * 10 + (c-'0');
			idx++;
		}
		
		if ( idx<sVersion.length() && sVersion.charAt(idx)=='.' ){
			idx++;
			
			while (idx<sVersion.length() && (c=sVersion.charAt(idx))>='0' && c<='9'){
				minor = minor * 10 + (c-'0');
				factor /= 10;
				idx++;
			}
		}
		
		return major + minor*factor;
	}
	
	/**
	 * Write mode for the current model
	 * 
	 * @return the write mode
	 */
	private WriteMode getWriteMode(){
		if (this.writeMode!=null)
			return this.writeMode;
		
		this.writeMode = this.mModel.getWriteMode();
		
		if (this.writeMode==WriteMode.ASUS_OLED_AUTO){
			if (Monitor.getConfig().getb("asus_oled.compatibility", false)){ //$NON-NLS-1$
				System.out.println("asus_oled compatibility mode forced by configuration"); //$NON-NLS-1$
				this.writeMode = WriteMode.ASUS_OLED_003;
			}
			else
			if (Monitor.getConfig().getb("asus_oled.enhanced", false)){ //$NON-NLS-1$
				System.out.println("asus_oled enhanced mode forced by configuration"); //$NON-NLS-1$
				this.writeMode = WriteMode.ASUS_OLED_004;
			}
			else{
				final String sVersion = Utils.getContents("/sys/class/asus_oled/version"); //$NON-NLS-1$
				
				System.out.println("asus_oled version from /sys/class/asus_oled/version : "+sVersion); //$NON-NLS-1$
	
				float version = getVersion(sVersion);
				
				System.out.println("numeric version: "+version); //$NON-NLS-1$
				
				if (version - 0.04f > -1E-5){
					System.out.println("writing to asus_oled in enhanced mode"); //$NON-NLS-1$
					this.writeMode = WriteMode.ASUS_OLED_004;
				}
				else{
					System.out.println("writing to asus_oled in compatibility mode"); //$NON-NLS-1$
					this.writeMode = WriteMode.ASUS_OLED_003;
				}
			}
		}
		
		return this.writeMode;
	}
	
	/**
	 * Test method 
	 * 
	 * @param args ignored
	 */
	public static void main(final String[] args) {
		//final Canvas c = new Canvas(Utils.getModel());
		final Canvas c = new Canvas(Model.G1);

		final Graphics2D graphics = c.getGraphics();

		graphics.setColor(Color.BLACK);
		graphics.fillRect(0, 0, c.W, c.H);
		
		graphics.setColor(Color.WHITE);		
		graphics.drawLine(10, 10, 200, 30);
		
		graphics.drawString("CPU", 10, 10); //$NON-NLS-1$
		
		c.writeImage();
	}
}
