/**
 * DBUS application enumerator
 */
package leds;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.StringTokenizer;

/**
 * @author costing
 * @since Jan 31, 2009
 */
public class DBUSApp implements KDEApp {
	
	/**
	 * System account name
	 */
	private final String sAccount;
	
	/**
	 * Session ID
	 */
	private final String sSession;
	
	/**
	 * Application name
	 */
	private final String sApp;
	
	/**
	 * Initialize the fields
	 * 
	 * @param account account name
	 * @param sessionID session id
	 * @param appName application name
	 */
	private DBUSApp(final String account, final String sessionID, final String appName){
		this.sAccount = account;
		this.sSession = sessionID;
		this.sApp = appName;
	}
	
	/**
	 * Call a function on this application
	 * 
	 * @param module application module
	 * @param function function name
	 * @param args arguments
	 * @return output
	 */
	public String execute(final String module, final String function, final String... args){
		final String[] commandLine = new String[6 + args.length];
		
		commandLine[0] = "utils/dbus/dbus_query.sh"; //$NON-NLS-1$
		commandLine[1] = this.sAccount;
		commandLine[2] = this.sSession;
		commandLine[3] = this.sApp;
		commandLine[4] = module;
		commandLine[5] = function;
		
		for (int i=0; i<args.length; i++)
			commandLine[6 + i] = args[i];
		
		return Utils.getProgramOutput(commandLine);
	}
	
	/**
	 * When was the application cache last generated
	 */
	private static long lLastUpdate = 0;
	
	/**
	 * Cached list of applications
	 */
	private static Collection<DBUSApp> appCache = null;
	
	/**
	 * Get the application with this name
	 * 
	 * @param name application name
	 * @return the application wrapper, or <code>null</code>
	 */
	public static DBUSApp getApplication(final String name){
		for (DBUSApp app: getApplications()){
			if (app.sApp.equals(name))
				return app;
		}
		
		return null;
	}
	
	/**
	 * Get DBUS-based applications we have access to
	 * 
	 * @return all the applications we can access
	 */
	public static synchronized final Collection<DBUSApp> getApplications(){
		if (System.currentTimeMillis() - lLastUpdate > 1000*5){
			appCache = reallyGetApplications();
			
			lLastUpdate = System.currentTimeMillis() + (appCache.size() > 0 ? 0 : 1000*15);	// less often queries if no DBUS app is around
		}
		
		return appCache;
	}

	/**
	 * Actually execute the program when necessary
	 * 
	 * @return list of applications
	 */
	private static final Collection<DBUSApp> reallyGetApplications(){
		final ArrayList<DBUSApp> apps = new ArrayList<DBUSApp>();
		
		final String sApps = Utils.getProgramOutput(new String[]{"utils/dbus/dbus_apps.sh"}); //$NON-NLS-1$
		
		if (sApps==null || sApps.length()==0)
			return apps;
		
		BufferedReader br = new BufferedReader(new StringReader(sApps));
		
		String sLine;
		
		try {
			while ( (sLine=br.readLine())!= null ){
				final StringTokenizer st = new StringTokenizer(sLine);
				
				try{
					apps.add(new DBUSApp(st.nextToken(), st.nextToken(), st.nextToken()));
				}
				catch (Exception e){
					// ignore
				}
			}
		}
		catch (IOException e) {
			// ignore
		}
		
		return apps;
	}
	
}
