/**
 * DCOPSession
 */
package leds;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * Wrapper around a DCOP session (identified by a system account and a DCOP server socket from the users' home directory).
 * 
 * @author costing
 * @since Dec 30, 2008
 */
public class DCOPSession {
	/**
	 * Account name
	 */
	private final String sUsername;
	
	/**
	 * Session path
	 */
	private final String sSession;
	
	/**
	 * Package protected constructor (built by {@link #listSessions()}) 
	 * 
	 * @param username user name
	 * @param session session path
	 * @see Utils#getDCOPApp(String)
	 */
	DCOPSession(final String username, final String session){
		this.sUsername = username;
		this.sSession = session;
	}
	
	/**
	 * Get all the applications that are running in this session. The list is generated at most every 5 seconds.
	 * 
	 * @return applications
	 */
	public Set<DCOPApp> getApplications(){
		final Set<DCOPApp> apps = new LinkedHashSet<DCOPApp>();
		
		final String sDCOP = Utils.getDCOPPath();
		
		if (sDCOP==null)
			return apps;
		
		final String[] commandLine = new String[] {sDCOP, "--no-user-time", "--user", this.sUsername, "--session", this.sSession};  //$NON-NLS-1$//$NON-NLS-2$ //$NON-NLS-3$
		
		final String sOutput = Utils.getProgramOutput(commandLine);
		
		if (sOutput==null || sOutput.length()==0)
			return apps;
		
		final BufferedReader br = new BufferedReader(new StringReader(sOutput));
		
		String sLine;
		
		try {
			while ( (sLine = br.readLine())!=null ){
				final DCOPApp app = new DCOPApp(this, sLine.trim());
				
				apps.add(app);
			}
		}
		catch (IOException e) {
			// ignore
		}
		
		return apps;
	}
	
	/**
	 * Command line to list DCOP sessions
	 */
	private static final String[] LIST_SESSIONS_CMD = new String[] {"dcop", "--no-user-time", "--all-users", "--list-sessions" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
	
	/**
	 * Last running DCOP sessions. Generated at 30 seconds, if requested.
	 */
	private static Set<DCOPSession> sessionsCache = new LinkedHashSet<DCOPSession>();
	
	/**
	 * Last sessions check time
	 */
	private static long lLastSessionsCheck = 0;
	
	/**
	 * Log a warning first time 'dcop' cannot be executed.
	 */
	private static boolean bDCOPNotFoundNotified = false;
	
	/**
	 * Get all the sessions running on this machine, for all users (if there are enough privileges).
	 * 
	 * @return DCOP sessions
	 * @see Utils#getDCOPApp(String)
	 */
	public static synchronized Set<DCOPSession> listSessions(){
		if (System.currentTimeMillis() < lLastSessionsCheck + 1000*30)
			return sessionsCache;
		
		final String sDCOP = Utils.getDCOPPath();
		
		if (sDCOP==null){
			// 15 minutes
			lLastSessionsCheck = System.currentTimeMillis() + 1000*60*15; 

			return sessionsCache;
		}

		final Set<DCOPSession> sessions = new LinkedHashSet<DCOPSession>();
		
		LIST_SESSIONS_CMD[0] = sDCOP;
		
		final String sOutput = Utils.getProgramOutput(LIST_SESSIONS_CMD);
	
		if (sOutput==null){
			if (!bDCOPNotFoundNotified){
				System.out.println("Didn't find 'dcop' command line tool in $PATH, disabling DCOP support"); //$NON-NLS-1$
				
				bDCOPNotFoundNotified = true;
			}
			
			// dcop executable not found in PATH, will sleep for a long time before trying again
			sessionsCache = sessions;
			
			// 15 minutes
			lLastSessionsCheck = System.currentTimeMillis() + 1000*60*15; 
			
			return sessionsCache;
		}
		
		final BufferedReader br = new BufferedReader(new StringReader(sOutput));
		
		String sLine = null;
		
		String sUser = null;
		
		try {
			while ( (sLine=br.readLine()) != null ){
				if (!sLine.startsWith(" ") && sLine.endsWith(":")){  //$NON-NLS-1$//$NON-NLS-2$
					// we get the user from here
					StringTokenizer st = new StringTokenizer(sLine);
					
					sUser = null;
					
					while (st.hasMoreTokens()){
						String sTok = st.nextToken();
						
						if (sTok.startsWith("/")){ //$NON-NLS-1$
							sUser = getUser(sTok);
							
							if (sUser!=null)
								break;
							
							System.out.println("Unknown system account name for '"+sTok+"'"); //$NON-NLS-1$ //$NON-NLS-2$
						}
					}
				}
				else{
					// this is a session
					
					final String sTrim = sLine.trim();
					
					if (sUser!=null && sTrim.length()>0)
						sessions.add(new DCOPSession(sUser, sTrim));
				}
			}
		}
		catch (IOException e) {
			// ignore, cannot be
		}
		
		sessionsCache = sessions;
		
		lLastSessionsCheck = System.currentTimeMillis();
		
		return sessionsCache;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return this.sUsername+":"+this.sSession;  //$NON-NLS-1$
	}
	
	/**
	 * Get the system account of this session.
	 * 
	 * @return account name
	 */
	public String getAccount(){
		return this.sUsername;
	}
	
	/**
	 * Get the session path
	 * 
	 * @return session path
	 */
	public String getSession(){
		return this.sSession;
	}
	
	/**
	 * Cache with home directories to user accounts mapping
	 */
	private static final HashMap<String, String> homedirsToUsers = new HashMap<String, String>();
	
	/**
	 * Get the system account for a given home directory
	 * 
	 * @param sHomeDir home directory path
	 * @return user account
	 */
	private static synchronized String getUser(final String sHomeDir){
		String sUser = homedirsToUsers.get(sHomeDir);
		
		if (sUser!=null)
			return sUser;
		
		homedirsToUsers.clear();
		
		final String sPasswd = Utils.getContents("/etc/passwd"); //$NON-NLS-1$
		
		final BufferedReader br = new BufferedReader(new StringReader(sPasswd));
		
		String sLine;
		
		try {
			while ( (sLine=br.readLine()) != null ){
				final StringTokenizer st = new StringTokenizer(sLine, ":"); //$NON-NLS-1$
				
				final String sUsername = st.nextToken();
				
				for (int i=0; i<4; i++)
					st.nextToken();
				
				final String sPath = st.nextToken();
				
				homedirsToUsers.put(sPath, sUsername);
			}
		}
		catch (Exception e) {
			System.out.println("Cannot parse /etc/passwd: "+e); //$NON-NLS-1$
		}
		
		return homedirsToUsers.get(sHomeDir);
	}
}
