/**
 * Flash the fancy LEDs on events
 */
package leds;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Flash a led device with controllable frequency for a given time interval.
 * 
 * @author costing
 * @since Nov 23, 2008
 */
public class Flasher extends Thread {

	/**
	 * Full path to the controlling device file. For example:
	 * /sys/class/leds/asus::touchpad/brightness
	 */
	private final String sDevice;
	
	/**
	 * How frequently to change turn the light on or off
	 */
	private long lSleep = 500;
	
	/**
	 * When to stop flashing
	 */
	private long lExpire = 0;
	
	/**
	 * Create a thread for a device
	 * 
	 * @param deviceName full path to the device controlling virtual file
	 */
	private Flasher(final String deviceName){
		super("leds.Flasher("+deviceName+")"); //$NON-NLS-1$ //$NON-NLS-2$
		setDaemon(true);
		
		this.sDevice = deviceName;
	}
	
	@Override
	public void run(){
		while (true){
			synchronized (this.sDevice) {
				try {
					if (System.currentTimeMillis() >= this.lExpire) {
						this.sDevice.wait();
					}
				}
				catch (InterruptedException e) {
					// ignore
				}
			}
			
			if (System.currentTimeMillis() < this.lExpire){
				final int oldValue = read(this.sDevice);
				int value = oldValue;
			
				while (System.currentTimeMillis() < this.lExpire){
					value=1-value;
				
					write(this.sDevice, value);
					
					try {
						sleep(this.lSleep);
					}
					catch (InterruptedException e) {
						// ignore
					}
				}
				
				write(this.sDevice, oldValue);
			}
		}
	}
	
	/**
	 * Set the interval between turning the LEDs on and off
	 * 
	 * @param interval interval, in milliseconds
	 * @return old value
	 */
	public long setFlickerInterval(final long interval){
		final long ret = this.lSleep;
		
		this.lSleep = interval;
		
		return ret;
	}
	
	/**
	 * Start flickering, for the specified amount of time.
	 * 
	 * @param lInterval interval, in milliseconds
	 */
	public void flicker(final long lInterval){
		if (this.lExpire < System.currentTimeMillis() + lInterval){
			this.lExpire = System.currentTimeMillis() + lInterval;
			
			synchronized(this.sDevice){
				this.sDevice.notify();
			}
		}
	}
	
	/**
	 * Stop flickering
	 */
	public void stopFlickering(){
		this.lExpire = 0;
	}
	
	/**
	 * All the devices that are in use
	 */
	private static final ConcurrentHashMap<String, Flasher> instances = new ConcurrentHashMap<String, Flasher>();

	/**
	 * Get the instance that controls the given device
	 * 
	 * @param sDevice full path to the device controlling virtual path (eg. "/sys/class/leds/asus::touchpad/brightness")
	 * @return an instance of Flasher or <code>null</code> if the device is not accessible
	 */
	public static synchronized Flasher getInstance(final String sDevice){
		Flasher f = instances.get(sDevice);
		
		if (f!=null)
			return f;
		
		if (read(sDevice)>=0){
			f = new Flasher(sDevice);
			f.start();
			instances.put(sDevice, f);
			
			return f;
		}
		
		return null;
	}
	
	/**
	 * Read the value from the virtual file
	 * 
	 * @param sDevice device path
	 * @return value currently in use
	 */
	private static int read(final String sDevice){
		BufferedReader br = null;
		
		try{
			br = new BufferedReader(new FileReader(sDevice));
		
			final String s = br.readLine();
		
			if (s!=null)
				return Integer.parseInt(s);
		}
		catch (Throwable t){
			// ignore
		}
		finally{
			if (br!=null){
				try{
					br.close();
				}
				catch (IOException ioe){
					// ignore
				}
			}
		}
		
		return -1;
	}

	/**
	 * Change the value
	 * 
	 * @param sDevice device path
	 * @param value new value
	 */
	private static void write(final String sDevice, final int value){
		try{
			final PrintWriter pw = new PrintWriter(sDevice);
			pw.println(value);
			pw.flush();
			pw.close();
		}
		catch (Exception e){
			// ignore
		}
	}
	
}
