/**
 * 
 */
package leds;


/**
 * Write mode
 * 
 * @author costing
 * @since Dec 24, 2008
 */
enum WriteMode {
	/**
	 * If this value is encountered then the code will try to look in /sys/class/asus_oled/version and figure out the version from there
	 */
	ASUS_OLED_AUTO,
	
	/**
	 * Old style write mode, supported by asus_oled until 0.03. The text to send to the driver is very long (one byte/pixel). 
	 */
	ASUS_OLED_003,
	
	/**
	 * New style write mode, from asus_oled 0.04 up. The chart is more efficient since it packs 8 pixels/byte.
	 */
	ASUS_OLED_004;
}

/**
 * Describe the various OLED displays that we can handle
 * 
 * @author costing
 * @since Dec 12, 2008
 */
public enum Model {
	/**
	 * Constant that shows that we are on a Asus G50 model (USB ID: 0b05:175b, size: 256x32 pixels).
	 * We might run on a new version of ASUS_OLED, or an old one. The evaluation will be done at run time.
	 */
	G1(128, 32, WriteMode.ASUS_OLED_003),
	
	/**
	 * Constant that shows that we are on a Asus G1 model (USB ID: 0b05:1726, size: 128x32 pixels)
	 * We are sure that we run on a newer version of the driver since the old one doesn't support this device.
	 */
	G50(256, 32, WriteMode.ASUS_OLED_004);
	
	/**
	 * Width of the display, in pixels
	 */
	private final int iWidth;
	
	/**
	 * Height of the display, in pixels
	 */
	private final int iHeight;
	
	/**
	 * What method to use
	 */
	private final WriteMode writeMode;
	
	/**
	 * @param width width, in pixels
	 * @param height height, in pixels
	 * @param mode write mode supported by the device
	 */
	private Model(final int width, final int height, final WriteMode mode){
		this.iWidth = width;
		this.iHeight = height;
		this.writeMode = mode;
	}
	
	/**
	 * Get the width of the display on this model
	 * 
	 * @return width, in pixels
	 */
	public int getWidth(){
		return this.iWidth;
	}
	
	/**
	 * Get the height of the display on this model
	 * 
	 * @return height, in pixels
	 */
	public int getHeight(){
		return this.iHeight;
	}
	
	/**
	 * Get the write mode supported by this device
	 * 
	 * @return write mode
	 */
	public WriteMode getWriteMode(){
		return this.writeMode;
	}
}
