package leds;

import java.awt.Font;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Observable;
import java.util.Observer;
import java.util.StringTokenizer;
import java.util.concurrent.atomic.AtomicBoolean;

/**
 * Gather monitoring information from the system, fill the image with the data and send it to the driver. The
 * application will also listen on localhost:1196 for commands. For now it can understand "pause" and "resume" commands.
 * These commands are sent by the ACPI event handlers included in this project (utils/ folder).<br>
 * <br>
 * The application is driven by the configuration file "config.properties". External modules can be added at any time,
 * you just have to implement the {@link leds.Module} interface (or extend {@link leds.modules.BaseModule}).
 * 
 * @author costing
 * @since Oct 7, 2008
 * @see Module
 */
public class Monitor implements Observer {

	/**
	 * Flag telling if the application is paused or not.
	 */
	static volatile boolean bPause = false;

	/**
	 * This internal class holds the text lines to display together with their expiration times
	 * 
	 * @author costing
	 * @since Oct 12, 2008
	 */
	private static final class TextLine {
		/**
		 * Text line to display
		 */
		public final String sLine;
	
		/**
		 * When it is going to expire
		 */
		public final long lExpires;
		
		/**
		 * Create an object with default expiration time (15 seconds from the moment of creation)
		 * 
		 * @param text text to display
		 */
		public TextLine(final String text){
			this(text, config.getl("notifications.expire", 1000*15)); //$NON-NLS-1$
		}
		
		/**
		 * Create an object with explicit text and expiration time
		 * 
		 * @param text text to display
		 * @param expires milliseconds relative to "now"
		 */
		public TextLine(final String text, final long expires){
			this.sLine = text;
			this.lExpires = System.currentTimeMillis() + expires;
		}
	}
	
	/**
	 * The list of text lines to display, with the expiration times for each
	 */
	static final List<TextLine> textLines = new LinkedList<TextLine>();
	
	/**
	 * For checking if some text was received during the current screen drawing 
	 */
	static final AtomicBoolean textReceived = new AtomicBoolean(false);
	
	/**
	 * Make the screen resume faster after lid open
	 */
	private static final Object oPauseLock = new Object();
	
	/**
	 * Serve a connection on localhost
	 * 
	 * @author costing
	 * @since Oct 12, 2008
	 */
	private static class TCPWorker extends Thread {
		/**
		 * The socket that is treated
		 */
		private final Socket socket;

		/**
		 * Initialize the worker with the socket
		 * 
		 * @param s socket to communicate
		 */
		public TCPWorker(final Socket s) {
			super("Worker"); //$NON-NLS-1$

			setDaemon(true);

			this.socket = s;
		}

		/**
		 * Read line by line and execute the commands provided
		 */
		@Override
		public void run() {
			try {
				this.socket.setSoTimeout(1000);
			}
			catch (Exception e) {
				// ignore
			}

			try {
				executeCommands(new InputStreamReader(this.socket.getInputStream(), "UTF-8")); //$NON-NLS-1$
			}
			catch (Throwable t) {
				// ignore
			}
			finally{
				try{
					this.socket.close();
				}
				catch (IOException ioe){
					// ignore
				}
			}
		}
	}
	
	/**
	 * Execute all the commands in this stream of lines
	 * 
	 * @param r any Reader
	 */
	static void executeCommands(final Reader r){
		final BufferedReader br = new BufferedReader(r);
		
		String sCMD;

		boolean bAdded = false;
		
		try{
			while ((sCMD = br.readLine()) != null) {
				bAdded = executeCommand(sCMD.trim()) || bAdded;
			}
		}
		catch (IOException ioe){
			// ignore
		}
		finally{
			try {
				br.close();
			}
			catch (IOException e) {
				// ignore
			}
		}
		
		synchronized (textLines){
			if (bAdded){
				textReceived.set(true);
				
				textLines.notifyAll();
			}
		}
	}
		
	/**
	 * Execute the commands from one line
	 * 
	 * @param sCMD command to execute
	 * @return true if some text line was added
	 */
	private static boolean executeCommand(final String sCMD) {
		boolean bReturn = false;
		
		if (sCMD.equals("pause")) //$NON-NLS-1$
			bPause = true;
		else
		if (sCMD.equals("resume")){ //$NON-NLS-1$
			bPause = false;
		
			synchronized (oPauseLock){
				oPauseLock.notifyAll();
			}
		}
		else
		if (sCMD.startsWith("text")) //$NON-NLS-1$
			synchronized(textLines){
				if (config.getb("notifications.purge_old", true)){ //$NON-NLS-1$
					while (textLines.size() >= 3)
						textLines.remove(0);
				}
				
				// put a reasonable limit on the buffer size
				if (textLines.size() < 32){
					textLines.add(new TextLine(sCMD.substring(4).trim()));

					bReturn = textLines.size() <= 3;
				}
			}
		
		if (bReturn)
			flickerLed();
		
		return bReturn;
	}

	/**
	 * Flash one of the fancy blue leds when a message has been received
	 */
	static void flickerLed(){
		final String sDevice = config.gets("flicker_on_new_message").trim(); //$NON-NLS-1$
		
		if (sDevice.length()>0){
			final Flasher f = Flasher.getInstance(sDevice);
			
			if (f!=null){
				f.setFlickerInterval(config.getl("flicker_on_new_message.how_often", 500)); //$NON-NLS-1$
				f.flicker(config.getl("flicker_on_new_message.for_how_long", 15000)); //$NON-NLS-1$
			}
		}
	}
	
	/**
	 * UDP server listening by default on 127.0.0.1:1196/UDP
	 * 
	 * @author costing
	 * @since Jan 11, 2009
	 */
	private static class UDPServer extends Thread {
		/**
		 * Simple default constructor
		 */
		public UDPServer(){
			super("UDPServer");  //$NON-NLS-1$
			
			setDaemon(true);
		}
		
		/* (non-Javadoc)
		 * @see java.lang.Thread#run()
		 */
		@Override
		public void run() {
			final DatagramSocket serverSocket;
			
			final int port = config.geti("server.udp.port", 1196); //$NON-NLS-1$
			final String address = config.gets("server.udp.address", "127.0.0.1");   //$NON-NLS-1$//$NON-NLS-2$
			
			try{
				serverSocket = new DatagramSocket(port, InetAddress.getByName(address));  
			}
			catch (Exception e){
				System.err.println("Cannot bind UDP server on "+address+':'+port); //$NON-NLS-1$
				return;
			}

			// will take only the first 256 bytes from the packet, we cannot display too much anyway
			final byte[] buff = new byte[256];
			
			final DatagramPacket packet = new DatagramPacket(buff, buff.length);
			
			while (true){
				try{
					serverSocket.receive(packet);
						
					final int len = packet.getLength();
						
					if (len>0){
						final String sContent = new String(buff, 0, len, "UTF8"); //$NON-NLS-1$
							
						executeCommands(new StringReader(sContent));
					}
				}
				catch (Exception e){
					// ignore
				}
			}
		}
	}
	
	/**
	 * Server thread listening for commands on localhost:1196
	 * 
	 * @author costing
	 * @since Oct 12, 2008
	 */
	private static class Server extends Thread {

		/**
		 * Default constructor
		 */
		public Server() {
			super("Server"); //$NON-NLS-1$

			setDaemon(true);
		}

		/**
		 * Infinite loop waiting for client connections
		 */
		@Override
		public void run() {
			final ServerSocket ss;
			
			final int port = config.geti("server.tcp.port", 1196);//$NON-NLS-1$
			final String address = config.gets("server.tcp.address", "127.0.0.1");//$NON-NLS-1$ //$NON-NLS-2$
			
			try {
				ss = new ServerSocket(port, 10, InetAddress.getByName(address));
			}
			catch (Exception e){
				System.err.println("Cannot bind TCP server on "+address+':'+port); //$NON-NLS-1$
				return;
			}
			
			while (true) {
				try{
					final Socket s = ss.accept();

					(new TCPWorker(s)).start();
				}
				catch (Exception e){
					// ignore
				}
			}
		}

	}
	
	/**
	 * Simple font
	 */
	private static final Font fDefault = new Font("Courier New", Font.PLAIN, 11); //$NON-NLS-1$

	/**
	 * The canvas used to talk to the device
	 */
	private static Canvas canvas;
	
	/**
	 * Configuration options read from ./config.properties
	 */
	static final ExtProperties config = new ExtProperties(".", "config"); //$NON-NLS-1$ //$NON-NLS-2$
	
	/**
	 * List of active modules
	 */
	private static final ArrayList<Module> modules = new ArrayList<Module>();
	
	/**
	 * Alternative modules to execute in case the main module fails
	 */
	private static final HashMap<Module, ArrayList<Module>> alternatives = new HashMap<Module, ArrayList<Module>>();
	
	/**
	 * Make default constructor unusable by anybody else
	 */
	private Monitor(){
		reloadConfig();
	}	
	
	/**
	 * Get the configuration options
	 * 
	 * @return options
	 */
	public static ExtProperties getConfig(){
		return config;
	}
	
	/**
	 * Call a module to draw. If it fails, call all it's alternative modules, recursively.
	 * 
	 * @param m module to draw
	 * @param c canvas to draw on
	 */
	private static void drawModule(final Module m, final Canvas c){
		boolean bOk = false;
		
		try{
			bOk = m.draw(c);
		}
		catch (Exception e){
			// ignore
		}
		
		if (!bOk){
			final ArrayList<Module> alt = alternatives.get(m);
			
			if (alt!=null){
				for (Module mAlt: alt)
					drawModule(mAlt, c);
			}
		}
	}

	/**
	 * Main loop of the application
	 * 
	 * @param args ignored
	 */
	public static void main(final String args[]) {

		canvas = new Canvas();
		
		// init display with something
		canvas.drawString("asusg50oled.sf.net", canvas.W / 2, 20, Canvas.ALIGN_CENTER); //$NON-NLS-1$

		if (!canvas.writeImage()) {
			System.err.println("Sorry, cannot find the asus_oled entry in /sys/class/asus_oled/. Please make sure you did all the following:\n- grabbed the latest version of asus_oled from SVN\n- executed make and make install for asus_oled\n- rmmod usbhid; modprobe asus_oled; modprobe usbhid."); //$NON-NLS-1$

			return;
		}
		
		config.setAutoReload(config.getl("config.check_changes", 0)); //$NON-NLS-1$
		config.addObserver(new Monitor());

		// start listening for commands
		
		if (config.getb("server.start", true)){ //$NON-NLS-1$
			final Server server = new Server();
			server.start();
			
			final UDPServer udpserver = new UDPServer();
			udpserver.start();
		}

		final AtomicBoolean bStop = new AtomicBoolean(false);

		// register close event to clear the display on shutdown
		Runtime.getRuntime().addShutdownHook(new Thread() {
			@Override
			public void run() {
				bStop.set(true);

				new Canvas().write();
			}
		});

		try {
			Thread.sleep(2000);
		}
		catch (Exception e) {
			// ignore
		}

		boolean bFirstPause = true;
		
		while (!bStop.get()) {
			if (bPause) {
				if (bFirstPause) {
					// clear the display only first time after the "pause" command was received
					canvas.fill(false);
					canvas.write();

					bFirstPause = false;
				}

				try {
					// wait for the lid to be opened.
					synchronized (oPauseLock){
						oPauseLock.wait(60000);
					}
				}
				catch (Exception e) {
					// ignore
				}

				continue;
			}

			final long lStarted = System.currentTimeMillis();
			
			bFirstPause = true;

			canvas.clearImage();
			canvas.getGraphics().setFont(fDefault);
			
			synchronized(textLines){
				// remove expired messages
				if (textLines.size()>0){
					final Iterator<TextLine> it = textLines.iterator();
				
					final long lNow = System.currentTimeMillis();
				
					while (it.hasNext()){
						final TextLine t = it.next();
					
						if (t.lExpires < lNow)
							it.remove();
						else
							break;
					}
				}
			}
			
			final int iTextLines = drawTextLines();
			
			if (iTextLines==0){
				synchronized (modules){
					for (Module m: modules){
						drawModule(m, canvas);
					}
				}
			}

			// we have finished drawing, transfer pixels to the internal buffer
			canvas.copyImage(canvas.getImage());

			if (iTextLines>0 && config.getb("invert_on_new_message", true)){ //$NON-NLS-1$
				// when there is a notification on display, invert the contents to make the event more visible
				canvas.invert();
			}
			
			if (!bStop.get()) {
				canvas.write();

				final long lSleepTime = config.geti("redraw_interval", 3000) - (System.currentTimeMillis() - lStarted);  //$NON-NLS-1$
				
				synchronized (textLines){
					try {
						// sleep only if some text was not received during the current drawing
						if (!textReceived.get() && lSleepTime > 0){
							// redraw the content every few seconds or when a new message is received and should be displayed
							textLines.wait(lSleepTime);
						}
					}
					catch (InterruptedException e) {
						// nothing bad, continue
					}
				}
			}
		}
	}
	
	/**
	 * Text lines positions
	 */
	private static int LINES_Y[] = {8, 18, 28};
	
	/**
	 * Take at most the first three lines of text and put them on the screen 
	 * 
	 * @return number of text lines displayed
	 */
	private static int drawTextLines(){
		int iLines = 0;
		
		int iMaxLines = Math.min(config.geti("notifications.lines", 3), 3); //$NON-NLS-1$
		
		synchronized (textLines){
			if (textLines.size() > 0){
				iLines = textLines.size();
				
				if (iLines>iMaxLines)
					iLines = iMaxLines;
			
				for (int i=0; i<iLines; i++){
					final String sLine = textLines.get(i).sLine;
										
					canvas.drawString(sLine, 0, LINES_Y[i]);
				}
			}
		}
		
		textReceived.set(false);
		
		return iLines;
	}

	/**
	 * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
	 */
	@Override
	public void update(final Observable o, final Object arg) {
		reloadConfig();
	}

	/**
	 * React on configuration file changes
	 */
	private static void reloadConfig(){
		loadModules();
	}
	
	/**
	 * Parse the alternatives string
	 * 
	 * @param m module for which the alternatives are added
	 * @param sAlternatives alternative modules to run
	 */
	private static void loadAlternatives(final Module m, final String sAlternatives){
		final ArrayList<Module> alt = new ArrayList<Module>();
		
		String sTemp = sAlternatives.trim();
		
		while (sTemp.length()>0){
			int idx1 = sTemp.indexOf('|');
			int idx2 = sTemp.indexOf(';');
			
			final Module mAlt;
			
			if (idx1<0 && idx2<0){
				// the only thing here is a simple class name
				mAlt = loadModule(sTemp.trim());
				sTemp = ""; //$NON-NLS-1$
			}
			else if (idx1<0 || idx2 < idx1){
				// next symbol is ';', so we load just the current module, leaving the rest for next iteration
				mAlt = loadModule(sTemp.substring(0, idx2).trim());
				sTemp = sTemp.substring(idx2+1).trim();
			}
			else{
				// next symbol is '|', everything that comes after is the alternative to this module
				mAlt = loadModule(sTemp.substring(0, idx1).trim());
				
				if (mAlt!=null){
					loadAlternatives(mAlt, sTemp.substring(idx1+1).trim());
				}
				
				sTemp = ""; //$NON-NLS-1$
			}
			
			if (mAlt!=null){
				alt.add(mAlt);
			}
		}
		
		if (alt.size()>0){
			alternatives.put(m, alt);
		}
	}
	
	/**
	 * Load a module
	 * 
	 * @param sModule module to load (and possibly its alternatives)
	 * @return the module instance, of <code>null</code> if the module could not be loaded
	 */
	private static Module loadModule(final String sModule){
		if (sModule.length()==0)
			return null;
		
		String sClass = sModule;
		String sAlternatives = ""; //$NON-NLS-1$
		
		if (sClass.indexOf('|')>0){
			sClass = sModule.substring(0, sModule.indexOf('|'));
			sAlternatives = sModule.substring(sModule.indexOf('|')+1);
		}
		
		Module m = null;
		
		try{
			m = (Module) Class.forName(sClass).newInstance();
			m.init(config, config.geti(sClass+".x", 0), config.geti(sClass+".y", 0));  //$NON-NLS-1$//$NON-NLS-2$
			
			if (sAlternatives.length()>0){
				loadAlternatives(m, sAlternatives);
			}
		}
		catch (Throwable t){
			System.err.println("Cannot load module '"+sModule+"' because "+t+" ("+t.getMessage()+")");   //$NON-NLS-1$//$NON-NLS-2$//$NON-NLS-3$ //$NON-NLS-4$
		}
		
		return m;
	}
	
	/**
	 * Load modules as described by the configuration tag "modules"
	 */
	private static void loadModules(){
		synchronized (modules){
			modules.clear();
			alternatives.clear();
			
			final StringTokenizer st = new StringTokenizer(config.gets("modules", ""), ", \t"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			
			while (st.hasMoreTokens()){
				String sModule = st.nextToken();
				
				Module m = loadModule(sModule);
				
				if (m!=null)
					modules.add(m);
			}
		}
	}
	
}
