package leds;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.util.Date;
import java.util.HashMap;
import java.util.StringTokenizer;

/**
 * Utility class for extracting various monitoring information from /proc files.
 * 
 * @author costing
 * @since Oct 7, 2008
 */
public class Utils {

	/**
	 * Get the entire contents of a file.
	 * 
	 * @param sFile file name
	 * @return contents as String, or <code>null</code> if there was a problem reading the file.
	 */
	public static final String getContents(final String sFile) {
		BufferedReader br = null;

		try {
			final StringBuilder sb = new StringBuilder();

			br = new BufferedReader(new FileReader(sFile));

			String sLine;

			while ((sLine = br.readLine()) != null) {
				if (sb.length()>0)
					sb.append('\n');
				
				sb.append(sLine);
			}

			return sb.toString();
		}
		catch (Throwable t) {
			return null;
		}
		finally {
			try {
				if (br!=null)
					br.close();
			}
			catch (Exception e) {
				// ignore
			}
		}
	}

	/**
	 * Old set of counters for the CPU usage
	 */
	private static Values cpuOld = null;

	/**
	 * Get the CPU utilization since the last call of this function. The returned map will contain the following keys:<br>
	 * <ul>
	 *  <li><i>usr</i></li>
	 *  <li><i>nice</i></li>
	 *  <li><i>sys</i></li>
	 *  <li><i>idle</i></li>
	 *  <li><i>iowait</i></li>
	 *  <li><i>int</i></li>
	 *  <li><i>softint</i></li>
	 *  <li><i>steal</i></li>
	 * </ul>
	 * 
	 * @return CPU utilization components
	 */
	public static Values getCPU() {
		String sCPU = getContents("/proc/stat"); //$NON-NLS-1$

		sCPU = sCPU.substring(sCPU.indexOf(" "), sCPU.indexOf("\n")).trim();  //$NON-NLS-1$//$NON-NLS-2$

		final StringTokenizer st = new StringTokenizer(sCPU);

		final Values cpuNew = new Values();

		if (st.hasMoreTokens())
			cpuNew.put("usr", Double.valueOf(st.nextToken())); //$NON-NLS-1$
		if (st.hasMoreTokens())
			cpuNew.put("nice", Double.valueOf(st.nextToken()));//$NON-NLS-1$
		if (st.hasMoreTokens())
			cpuNew.put("sys", Double.valueOf(st.nextToken()));//$NON-NLS-1$
		if (st.hasMoreTokens())
			cpuNew.put("idle", Double.valueOf(st.nextToken()));//$NON-NLS-1$
		if (st.hasMoreTokens())
			cpuNew.put("iowait", Double.valueOf(st.nextToken()));//$NON-NLS-1$
		if (st.hasMoreTokens())
			cpuNew.put("int", Double.valueOf(st.nextToken()));//$NON-NLS-1$
		if (st.hasMoreTokens())
			cpuNew.put("softint", Double.valueOf(st.nextToken()));//$NON-NLS-1$
		if (st.hasMoreTokens())
			cpuNew.put("steal", Double.valueOf(st.nextToken()));//$NON-NLS-1$

		if (cpuOld != null) {
			Values ret = cpuNew.substract(cpuOld);

			double sum = 0;

			for (double d : ret.values())
				sum += d;

			ret = ret.div(sum);

			cpuOld = cpuNew;

			return ret;
		}
		
		cpuOld = cpuNew;

		return null;
	}

	/**
	 * Old set of counters for the network traffic
	 */
	private static Values netOldIO;

	/**
	 * Get the network traffic information. The mapping will contain for each network interface the following values:<br>
	 * <ul>
	 *  <li><i>[name]_IN</i> - received bytes divided by time since last measurement, so B/s</li>
	 *  <li><i>[name]_OUT</i> - sent bytes, divided by time since last measurement</li>
	 *  <li><i>[name]_IN_cnt</i> - number of received packets, divided by time since last measurement, so packets/s</li>
	 *  <li><i>[name]_OUT_cnt</i> - number of sent packets, divided by time since last measurement</li>
	 * </ul>
	 * Additionally all the values are summed up under a virtual interface, "TOTAL".
	 * 
	 * @return network monitoring
	 */
	public static Values getNetIO() {
		final Values netNewIO = new Values();

		final BufferedReader br = new BufferedReader(new StringReader(getContents("/proc/net/dev"))); //$NON-NLS-1$

		String sLine;

		netNewIO.put("time", Double.valueOf(System.currentTimeMillis() / 1000.)); //$NON-NLS-1$

		try {
			double dTotalINBytes = 0;
			double dTotalINPackets = 0;
			double dTotalOUTBytes = 0;
			double dTotalOUTPackets = 0;

			while ((sLine = br.readLine()) != null) {
				final int idx = sLine.indexOf(":"); //$NON-NLS-1$

				if (idx < 0)
					continue;

				final String sIFName = sLine.substring(0, idx).trim();

				StringTokenizer st = new StringTokenizer(sLine.substring(idx + 1).trim());

				final double dINBytes = Double.parseDouble(st.nextToken());
				final double dINPackets = Double.parseDouble(st.nextToken());
				netNewIO.put(sIFName + "_IN", Double.valueOf(dINBytes)); //$NON-NLS-1$
				netNewIO.put(sIFName + "_IN_cnt", Double.valueOf(dINPackets)); //$NON-NLS-1$

				for (int i = 0; i < 6; i++)
					st.nextToken();

				final double dOUTBytes = Double.parseDouble(st.nextToken());
				final double dOUTPackets = Double.parseDouble(st.nextToken());
				netNewIO.put(sIFName + "_OUT", Double.valueOf(dOUTBytes)); //$NON-NLS-1$
				netNewIO.put(sIFName + "_OUT_cnt", Double.valueOf(dOUTPackets)); //$NON-NLS-1$

				dTotalINBytes += dINBytes;
				dTotalINPackets += dINPackets;
				dTotalOUTBytes += dOUTBytes;
				dTotalOUTPackets += dOUTPackets;
			}

			netNewIO.put("TOTAL_IN", Double.valueOf(dTotalINBytes)); //$NON-NLS-1$
			netNewIO.put("TOTAL_IN_cnt", Double.valueOf(dTotalINPackets)); //$NON-NLS-1$
			netNewIO.put("TOTAL_OUT", Double.valueOf(dTotalOUTBytes)); //$NON-NLS-1$
			netNewIO.put("TOTAL_OUT_cnt", Double.valueOf(dTotalOUTPackets)); //$NON-NLS-1$
		}
		catch (IOException ioe) {
			// ignore
		}

		if (netOldIO != null) {
			Values ret = netNewIO.substract(netOldIO);

			ret = ret.div(ret.getz("time")); //$NON-NLS-1$

			netOldIO = netNewIO;

			return ret;
		}
		
		netOldIO = netNewIO;

		return null;
	}

	/**
	 * Get load of the system. The mapping will have the following keys:<br>
	 * <ul>
	 *  <li><i>Load1</i></li>
	 *  <li><i>Load5</i></li>
	 *  <li><i>Load15</i></li>
	 * </ul>
	 * 
	 * @return load components 
	 */
	public static Values getLoad() {
		final Values ret = new Values();

		final StringTokenizer st = new StringTokenizer(getContents("/proc/loadavg")); //$NON-NLS-1$

		ret.put("Load1", Double.valueOf(st.nextToken())); //$NON-NLS-1$
		ret.put("Load5", Double.valueOf(st.nextToken())); //$NON-NLS-1$
		ret.put("Load15", Double.valueOf(st.nextToken())); //$NON-NLS-1$

		return ret;
	}

	/**
	 * Where do we get the battery info from?
	 * 0 = undetermined yet
	 * 1 = /sys/class/power_supply/BAT0
	 * 2 = /proc/acpi/battery/BAT0
	 */
	private static int iBatteryInfoSource = 0;
	
	/**
	 * Battery capacity, last time when it was checked
	 */
	private static int iBatteryCapacity = -1;
	
	/**
	 * Timestamp of the last battery capacity check. We only check the capacity once every 10 minutes, it doesn't change in time.
	 */
	private static long lBatteryCapacityLastCheck = 0;
	
	/**
	 * How often to check the battery capacity ?
	 */
	private static final long CHECK_BATTERY_CAPACITY = 1000*60*10;
	
	/**
	 * Get the battery status. The returned values will have the following keys:<br>
	 * <ul>
	 * 	<li><i>capacity</i></li>
	 *  <li><i>remaining</i></li>
	 *  <li><i>percentage</i></li>
	 * </ul>
	 * 
	 * @return battery status as {@link Values} mapping
	 */
	public static Values getBattery() {
		if (iBatteryInfoSource==0){
			String sTest = getContents("/sys/class/power_supply/BAT0/energy_full");  //$NON-NLS-1$
			
			if (sTest!=null && sTest.length()>0){
				System.out.println("Getting the battery info from /sys/class/power_supply/BAT0/{energy_full,energy_now}"); //$NON-NLS-1$
				iBatteryInfoSource = 1;
			}
			else{
				sTest = getContents("/proc/acpi/battery/BAT0/info"); //$NON-NLS-1$
				
				if (sTest!=null && sTest.length()>0){
					System.out.println("Getting the battery info from /proc/acpi/battery/BAT0/{info,state}"); //$NON-NLS-1$
					iBatteryInfoSource = 2;
				}
				else{
					System.out.println("I cannot read battery info from either /sys/class/power_supply/BAT0/ or /proc/acpi/battery/BAT0/. I'll try again later."); //$NON-NLS-1$
					return null;
				}
			}
		}

		int capacity = 0;
		int remaining = 0; 
		
		if (iBatteryInfoSource == 1){
			try{
				if (System.currentTimeMillis() - lBatteryCapacityLastCheck > CHECK_BATTERY_CAPACITY){
					final String sEnergyFull = getContents("/sys/class/power_supply/BAT0/energy_full"); //$NON-NLS-1$
					
					if (sEnergyFull == null || sEnergyFull.length() == 0){
						lBatteryCapacityLastCheck = 0;
						return null;
					}
					
					iBatteryCapacity = Integer.parseInt(sEnergyFull)/1000;
					lBatteryCapacityLastCheck = System.currentTimeMillis();
				}
				
				capacity = iBatteryCapacity; 
				
				final String sEnergyNow = getContents("/sys/class/power_supply/BAT0/energy_now");  //$NON-NLS-1$
				
				if (sEnergyNow == null || sEnergyNow.length() == 0){
					lBatteryCapacityLastCheck = 0;
					return null;
				}

				remaining = Integer.parseInt(sEnergyNow)/1000;
			}
			catch (Exception e){
				System.out.println("Cannot parse the contents of /sys/class/power_supply/BAT0/{energy_full,energy_now} : "+e+" ("+e.getMessage()+")");  //$NON-NLS-1$//$NON-NLS-2$ //$NON-NLS-3$				
				lBatteryCapacityLastCheck = 0;
				return null;
			}
		}
		else
		if (iBatteryInfoSource == 2){
			try{
				if (System.currentTimeMillis() - lBatteryCapacityLastCheck > CHECK_BATTERY_CAPACITY){
					final String sInfo = getContents("/proc/acpi/battery/BAT0/info"); //$NON-NLS-1$
					
					if (sInfo == null || sInfo.length() == 0){
						lBatteryCapacityLastCheck = 0;
						return null;
					}
	
					int idx = sInfo.indexOf("last full capacity"); //$NON-NLS-1$
					idx = sInfo.indexOf(":", idx) + 1; //$NON-NLS-1$
					int idx2 = sInfo.indexOf("mWh", idx); //$NON-NLS-1$
	
					String s = sInfo.substring(idx, idx2).trim();
	
					iBatteryCapacity = Integer.parseInt(s);
					lBatteryCapacityLastCheck = System.currentTimeMillis();
				}
				
				capacity = iBatteryCapacity;
				
				
				final String sState = getContents("/proc/acpi/battery/BAT0/state"); //$NON-NLS-1$
	
				if (sState == null || sState.length()==0){
					lBatteryCapacityLastCheck = 0;
					return null;
				}
	
				int idx = sState.indexOf("remaining capacity"); //$NON-NLS-1$
				idx = sState.indexOf(":", idx) + 1; //$NON-NLS-1$
				int idx2 = sState.indexOf("mWh", idx); //$NON-NLS-1$
	
				String s = sState.substring(idx, idx2).trim();
	
				remaining = Integer.parseInt(s);
			}
			catch (Exception e){
				System.out.println("Cannot parse the contents of /proc/acpi/battery/BAT0/{info,state} : "+e+" ("+e.getMessage()+")");  //$NON-NLS-1$//$NON-NLS-2$ //$NON-NLS-3$				
				lBatteryCapacityLastCheck = 0;
				return null;
			}
		}
		else {
			return null; // ?!
		}
		
		final Values ret = new Values();

		ret.put("capacity", Double.valueOf(Math.max(capacity, remaining))); //$NON-NLS-1$
		ret.put("remaining", Double.valueOf(remaining)); //$NON-NLS-1$		
		ret.put("percentage", Double.valueOf(Math.min((double) remaining / capacity, 1))); //$NON-NLS-1$

		return ret;
	}
	
	/**
	 * Where to read the CPU temperature from.
	 * 0 = yet unknown
	 * 1 = /sys/class/thermal/thermal_zone0/temp
	 * 2 = /proc/acpi/thermal_zone/THRM/temperature
	 */
	private static int iTemperatureSource = 0;
	
	/**
	 * CPU thermal zone info from /sys
	 */
	private static final String CPU_TEMP_SYS = "/sys/class/thermal/thermal_zone0/temp"; //$NON-NLS-1$
	
	/**
	 * CPU thermal zone info from /proc 
	 */
	private static final String CPU_TEMP_PROC = "/proc/acpi/thermal_zone/THRM/temperature"; //$NON-NLS-1$
	
	/**
	 * Value to show an error getting the CPU temperature. In this case the return value of {@link #getCPUTemperature()} should be discarded
	 */
	public static final int CPU_TEMP_ERROR = -1000;
	
	/**
	 * Get the CPU temperature.
	 * 
	 * @return CPU temperature, in &deg;C, or {@link #CPU_TEMP_ERROR} if the temperature cannot be read from the ACPI virtual file.
	 */
	public static int getCPUTemperature(){
		if (iTemperatureSource==0){
			String sTest = getContents(CPU_TEMP_SYS);
			
			if (sTest!=null && sTest.length()>0){
				System.out.println("Reading CPU temperature from "+CPU_TEMP_SYS); //$NON-NLS-1$
				iTemperatureSource = 1;
			}
			else{
				sTest = getContents(CPU_TEMP_PROC);
				
				if (sTest!=null && sTest.length()>0){
					System.out.println("Reading CPU temperature from "+CPU_TEMP_PROC); //$NON-NLS-1$
					iTemperatureSource = 2;
				}
				else{
					System.out.println("Cannot read the CPU temperature from either "+CPU_TEMP_SYS+" or "+CPU_TEMP_PROC); //$NON-NLS-1$ //$NON-NLS-2$
					return CPU_TEMP_ERROR;
				}
			}
		}
		
		if (iTemperatureSource==1){
			try{
				return Integer.parseInt(getContents(CPU_TEMP_SYS))/1000;
			}
			catch (Exception e){
				System.out.println("Cannot parse the CPU temperature from "+CPU_TEMP_SYS+" : "+e);  //$NON-NLS-1$//$NON-NLS-2$
				return CPU_TEMP_ERROR;
			}
		}
		
		if (iTemperatureSource==2){
			final String sTemp = getContents(CPU_TEMP_PROC);
			
			if (sTemp==null)
				return CPU_TEMP_ERROR;
			
			final StringTokenizer st = new StringTokenizer(sTemp);
			
			if (st.countTokens()<2)
				return CPU_TEMP_ERROR;
			
			st.nextToken();
			
			try{
				return Integer.parseInt(st.nextToken());
			}
			catch (Exception e){
				System.out.println("Cannot parse the CPU temperature from "+CPU_TEMP_PROC+" : "+e);  //$NON-NLS-1$//$NON-NLS-2$
				return CPU_TEMP_ERROR;
			}
		}
	
		return CPU_TEMP_ERROR;
	}

	/**
	 * Where to get the CPUs frequencies from?
	 * 0 = yet unknown
	 * 1 = /sys/devices/system/cpu/cpuX/cpufreq/scaling_cur_freq
	 * 2 = /proc/cpuinfo
	 */
	private static int iCPUFreqInfoSource = 0;
	
	/**
	 * Get the CPUs frequencies. The returned map will have as many CPU# entries as CPUs/cores, and the value for each 
	 * will be the current operating frequency.
	 * 
	 * @return CPUs freq
	 */
	public static Values getCPUFreq() {
		final Values ret = new Values();
		
		if (iCPUFreqInfoSource==0){
			String sTest = getContents("/sys/devices/system/cpu/cpu0/cpufreq/scaling_cur_freq"); //$NON-NLS-1$
			
			if (sTest!=null && sTest.length()>0){
				System.out.println("Reading CPU cores' frequencies from /sys/devices/system/cpu/cpuX/cpufreq/scaling_cur_freq"); //$NON-NLS-1$
				iCPUFreqInfoSource = 1;
			}
			else{
				sTest = getContents("/proc/cpuinfo"); //$NON-NLS-1$
				
				if (sTest!=null && sTest.length()>0){
					System.out.println("Reading CPU cores' frequencies from /proc/cpuinfo"); //$NON-NLS-1$
					iCPUFreqInfoSource = 2;
				}
				else{
					System.out.println("Cannot read CPU cores' frequencies from either /sys or /proc (?!)"); //$NON-NLS-1$
					return ret;
				}
			}
		}
		
		if (iCPUFreqInfoSource==1){
			for (int i=Runtime.getRuntime().availableProcessors()-1; i>=0; i--){
				try{
					final int freqMHz = Integer.parseInt(getContents("/sys/devices/system/cpu/cpu"+i+"/cpufreq/scaling_cur_freq"))/1000;  //$NON-NLS-1$//$NON-NLS-2$
					
					ret.put("CPU"+i, Double.valueOf(freqMHz)); //$NON-NLS-1$
				}
				catch (Exception e){
					System.out.println("Cannot parse CPU freq : "+e); //$NON-NLS-1$
				}
			}
		}
		else
		if (iCPUFreqInfoSource == 2){		
			final String sProc = getContents("/proc/cpuinfo"); //$NON-NLS-1$
	
			int idx = sProc.indexOf("cpu MHz"); //$NON-NLS-1$
	
			int cnt = 0;
	
			while (idx >= 0) {
				idx = sProc.indexOf(":", idx); //$NON-NLS-1$
	
				final String sValue = sProc.substring(idx + 1, sProc.indexOf("\n", idx)).trim(); //$NON-NLS-1$
	
				final double value = Double.parseDouble(sValue);
	
				ret.put("CPU" + cnt, Double.valueOf(value)); //$NON-NLS-1$
				cnt++;
	
				idx = sProc.indexOf("cpu MHz", idx); //$NON-NLS-1$
			}
		}

		return ret;
	}
	
	/**
	 * Get the output of the given command
	 * 
	 * @param commandLine program and command line arguments
	 * @return full output, or <code>null</code> if there was any problem running the program
	 */
	public static String getProgramOutput(final String[] commandLine){
		final boolean bDebug = Monitor.getConfig().getb("Utils.getProgramOutput.debug", false); //$NON-NLS-1$
		
		String sCommand = ""; //$NON-NLS-1$
		
		if (bDebug){
			System.out.print(new Date().toString());
			System.out.print(": executing command: '"); //$NON-NLS-1$
			
			for (int i=0; i<commandLine.length; i++){
				sCommand += (sCommand.length() > 0 ? " " : "") + commandLine[i];  //$NON-NLS-1$//$NON-NLS-2$
			}
			
			System.out.print(sCommand);
			System.out.println("'"); //$NON-NLS-1$
			System.out.flush();
		}
		
		final StringBuilder sb = new StringBuilder();
		
		try{
			final Process p = Runtime.getRuntime().exec(commandLine);
			
			p.getOutputStream().close();
			final BufferedReader br = new BufferedReader(new InputStreamReader(p.getInputStream()));
			
			String sLine;

			while ( (sLine=br.readLine())!=null ){
				if (sb.length()>0)
					sb.append('\n');
				
				sb.append(sLine);
			}

			br.close();
			
			p.getErrorStream().close();
			
			p.waitFor();
		}
		catch (Throwable t){
			if (bDebug){
				System.err.println("Cannot get program output  : "+t+" ("+t.getMessage()+")");  //$NON-NLS-1$//$NON-NLS-2$ //$NON-NLS-3$
				t.printStackTrace();
			}
			
			return null;
		}
		
		final String sOutput = sb.toString();
		
		if (bDebug && Monitor.getConfig().getb("Utils.getProgramOutput.debug.output", false)){ //$NON-NLS-1$
			System.out.print(new Date().toString());
			System.out.println(": output of '"+sCommand+"':"); //$NON-NLS-1$ //$NON-NLS-2$
			System.out.println(sOutput);
		}
		
		return sOutput;
	}

	/**
	 * Get the hdd parameters for a given device.
	 * 
	 * @param sDevice device name, usually "/dev/sda" or similar
	 * @return key-value mappings, similar to<br>
	 *                  <code>smartctl -A /dev/sda | grep -e "0x"  | awk '{print $2 " " $10}'</code>
	 */
	public static Values getHDDParameters(final String sDevice){
		final String sOutput = getProgramOutput(new String[]{"smartctl", "-A", sDevice}); //$NON-NLS-1$ //$NON-NLS-2$

		final Values ret = new Values();

		if (sOutput==null || sOutput.length()==0)
			return ret; 
		
		try{
			final BufferedReader br = new BufferedReader(new StringReader(sOutput));
			
			String sLine;
			
			while ( (sLine=br.readLine())!=null ){
				if (sLine.indexOf("0x") < 0) continue; //$NON-NLS-1$
				
				final StringTokenizer st = new StringTokenizer(sLine);
				
				if (st.countTokens()>=10){
					st.nextToken();
					
					final String sKey = st.nextToken();
					
					for (int i=0; i<7; i++) st.nextToken();
					
					try{
						final int i = Integer.parseInt(st.nextToken());
					
						ret.put(sKey, new Double(i));
					}
					catch (Exception e){
						System.err.println("Ignoring the smartctl value for '"+sKey+"' because value parsing to integer failed with "+e+" ("+e.getMessage()+")");   //$NON-NLS-1$//$NON-NLS-2$//$NON-NLS-3$ //$NON-NLS-4$
					}
				}
			}
		}
		catch (Throwable t){
			System.err.println("Cannot parse output of smartctl : "+t+" ("+t.getMessage()+")");  //$NON-NLS-1$//$NON-NLS-2$ //$NON-NLS-3$
			t.printStackTrace();
		}
		
		return ret;
	}
	
	/**
	 * Last replacement size
	 */
	private static volatile int	iOldReplacements1	= 2;

	/**
	 * Second last replacement size
	 */
	private static volatile int	iOldReplacements2	= 2;

	/**
	 * Replace a block of text with another text in a given String.
	 * 
	 * @param s original text
	 * @param sWhat what to search and replace
	 * @param sWith the new text to put in place
	 * @return the modified text
	 */
	public static String replace(final String s, final String sWhat, final String sWith) {
		if (s==null || sWhat==null || sWhat.length()==0)
			return s;

		StringBuilder sb = null;

		final int iWhatLen = sWhat.length();
		final int iWithLen = sWith.length();
		
		int iOld = 0;
		int i = 0;
		int iReplacements = 0;
		
		while ((i = s.indexOf(sWhat, iOld)) >= 0) {
			if (sb == null) {
				final int diff = (iWhatLen - iWithLen) * (iOldReplacements1 * 2 + iOldReplacements2 + 1) / 3;
				sb = new StringBuilder(s.length() + (diff > 0 ? diff : 0));
			}

			if (iOld<i)
				sb.append(s.substring(iOld, i));
			
			sb.append(sWith);
			
			iOld = i + iWhatLen;
			
			iReplacements++;
		}

		iOldReplacements2 = iOldReplacements1;
		iOldReplacements1 = iReplacements;

		if (sb != null) {
			sb.append(s.substring(iOld));
			return sb.toString();
		}
		
		// no replacements to do
		return s;
	}
	
	/**
	 * Convert a String to a boolean. It will take any hint at the beginning of the string
	 * (t, T, y, Y, 1 / f, F, n, N, 0) to return <code>true</code> / <code>false</code>.  
	 * 
	 * @param s string to convert
	 * @param bDefault default value to return in case the given string doesn't show up any of the predefined signs
	 * @return the boolean value
	 */
	public static boolean stringToBool(final String s, final boolean bDefault){
		if (s!=null && s.length()>0){
			final char c = s.charAt(0);
			
			if (c=='t' || c=='T' || c=='y' || c=='Y' || c=='1') return true;
			if (c=='f' || c=='F' || c=='n' || c=='N' || c=='0') return false;
		}
		
		return bDefault;
	}
	
	/**
	 * Cached model of the display
	 */
	private static Model model = null;
	
	/**
	 * Get the OLED model, looking in the configuration file for "G50=true" or "G1=true". If none is found then it assumes G50 platform.
	 * The value is inferred only once, at the first call. Afterwards it always keeps returning this initial value, assuming that the
	 * platform doesn't change on the fly during execution.
	 * 
	 * @return one of the {@link Model} enum values
	 */
	public static Model getModel(){
		if (model != null)
			return model;
		
		final ExtProperties config = Monitor.getConfig();
		
		// is the model forced from the configuration file to G50 (with G50=true) ?
		if (config.getb("G50", false)){ //$NON-NLS-1$
			System.out.println("ASUS G50 forced in the configuration file"); //$NON-NLS-1$
			model = Model.G50;
			return model;
		}
		
		// is the model forced from the configuration file to G50 (with G50=true) ?
		if (config.getb("G1", false)){ //$NON-NLS-1$
			System.out.println("ASUS G1 forced in the configuration file"); //$NON-NLS-1$
			model = Model.G1;
			return model;
		}
		
		if (model == null){
			// ok, let's see what hardware we have ...
			String lsusb = getProgramOutput(new String[]{"lsusb"}); //$NON-NLS-1$
			
			if (lsusb!=null){
				lsusb = lsusb.toLowerCase();
				if (lsusb.indexOf("0b05:175b")>=0){ //$NON-NLS-1$
					System.out.println("lsusb shows that we are running on a G50 device"); //$NON-NLS-1$
					model = Model.G50;
				}
				else
				if (lsusb.indexOf("0b05:1726")>=0){ //$NON-NLS-1$
					System.out.println("lsusb shows that we are running on a G1 device"); //$NON-NLS-1$
					model = Model.G1;
				}
				else{
					System.out.println("Cannot see any compatible OLED device in your lsusb output:\n"+lsusb); //$NON-NLS-1$
				}
			}
		}
		
		if (model == null){
			// everything failed, assuming some bogus device 
			model = Model.G50;	// default value
		}
		
		return model;
	}
	
	/**
	 * DCOP applications cache
	 */
	static final HashMap<String, DCOPApp> appCache = new HashMap<String, DCOPApp>(); 
	
	/**
	 * Flag to indicate some DCOP-based module calling us
	 */
	static boolean bDCOPInUse = false;
	
	/**
	 * Thread that lists all DCOP applications in background
	 */
	private static DCOPApplicationsFetcherThread appFetcherThread = null;
	
	/**
	 * Get the DCOP application that has this name.
	 * 
	 * @param applicationName application name
	 * @return the application, or <code>null</code> if the application is not running
	 */
	public static DCOPApp getDCOPApp(final String applicationName){
		bDCOPInUse = true;
		
		synchronized (appCache){
			if (appFetcherThread==null){
				appFetcherThread = new DCOPApplicationsFetcherThread();
				appFetcherThread.start();
			}
			
			return appCache.get(applicationName);
		}		
	}
	
	/**
	 * Update the cache of active DCOP applications every 10 seconds
	 * 
	 * @author costing
	 * @since Jan 29, 2009
	 */
	private static final class DCOPApplicationsFetcherThread extends Thread {
		/**
		 * Default constructor
		 */
		public DCOPApplicationsFetcherThread() {
			super("leds.Utils.DCOPApplicationsFetcherThread"); //$NON-NLS-1$
			setDaemon(true);
		}
		
		@Override
		public void run(){
			while (true){
				final HashMap<String, DCOPApp> apps = new HashMap<String, DCOPApp>();
				
				boolean bDCOPSessions = false;
				
				for (DCOPSession session: DCOPSession.listSessions()){
					bDCOPSessions = true;
					
					for (DCOPApp app: session.getApplications()){
						apps.put(app.getApplication(), app);
					}
				}
				
				synchronized (appCache){
					appCache.clear();
					appCache.putAll(apps);
				}
				
				final int sleepTime = 1000 * (bDCOPInUse ? (bDCOPSessions ? 5 : 10) : 60);
				
				bDCOPInUse = false;
				
				try{
					sleep(sleepTime);
				}
				catch (Exception e){
					// ignore
				}
			}
		}
	}
	
	/**
	 * Where is "dcop" executable?
	 */
	private static String dcopPath = null;
	
	/**
	 * Have we already tried to determine where "dcop" executable is on disk ?
	 */
	private static boolean dcopPathDetermined = false;
	
	/**
	 * Standard paths where "dcop" can be found
	 */
	private static final String[] dcopPaths = {
		"/usr/bin",			// Ubuntu, RedHat, Slackware //$NON-NLS-1$
		"/opt/kde3/bin",	// SuSE //$NON-NLS-1$
		"/usr/local/bin",	// Google //$NON-NLS-1$
		"/usr/kde/3.5/bin"	// Something else from Google //$NON-NLS-1$
	};
	
	/**
	 * Get the full path to "dcop" executable. Will look in several usual locations, but it can also be overriden by setting
	 * <i>Utils.dcop_path</i> in the configuration file.
	 * 
	 * @return full path to "dcop" or <code>null</code> if it cannot be found in any of the known folders
	 */
	public static synchronized String getDCOPPath(){
		if (dcopPathDetermined)
			return dcopPath;

		dcopPathDetermined = true;
		
		dcopPath = Monitor.getConfig().gets("Utils.dcop_path"); //$NON-NLS-1$
		
		if (dcopPath.length()>0){
			final File f = new File(dcopPath);
			
			if (f.exists() && f.isFile() && f.canExecute()){
				System.err.println("dcop: taken from configuration file, seems to be an executable all right"); //$NON-NLS-1$
				return dcopPath;
			}
			
			System.err.println("dcop: the configuration file setting doesn't seem to point to an executable"); //$NON-NLS-1$
		}
		
		dcopPath = getPath("dcop", dcopPaths); //$NON-NLS-1$
		
		if (dcopPath!=null){
			System.err.println("dcop: found in "+dcopPath); //$NON-NLS-1$
			dcopPath += "/dcop"; //$NON-NLS-1$
		}
		else{
			System.err.println("dcop: couldn't find anywhere on the system. If you have it, please set 'Utils.dcop_path' in the configuration file to point to it"); //$NON-NLS-1$
		}
		
		return dcopPath;
	}
	
	/**
	 * From the list of possible locations, pick the first where the given executable is found
	 * 
	 * @param executable executable to search for
	 * @param folders folders to look at
	 * @return the first folder where the executable is found, or <code>null</code> if none is found
	 */
	private static String getPath(final String executable, final String[] folders){
		for (String sPath: folders){
			final File f = new File(sPath, executable);
			
			if (f.exists() && f.isFile() && f.canExecute())
				return sPath;
		}
		
		return null;
	}
	
	/**
	 * Debug method
	 * 
	 * @param args ignored
	 */
	public static void main(String[] args) {
		System.err.println("Amarok: "+getDCOPApp("amarok"));  //$NON-NLS-1$//$NON-NLS-2$
		System.err.println("Kaffeine: "+getDCOPApp("kaffeine"));  //$NON-NLS-1$//$NON-NLS-2$
		
		//System.err.println(getDCOPApp("amarok").execute("player", "title"));  //$NON-NLS-1$ //$NON-NLS-2$//$NON-NLS-3$
	}
}
