/**
 * Amarok.java
 */
package leds.modules;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;

import leds.Canvas;
import leds.DBUSApp;
import leds.DCOPApp;
import leds.ExtProperties;
import leds.Utils;

/**
 * The Amarok module gets the current status of Amarok via DCOP calls. It will try to get the values for any user,
 * so that running the daemon as root will still display the active session that belongs to a normal user.
 * 
 * @author costing
 * @since Dec 21, 2008
 */
public class Amarok extends BaseModule {
	/**
	 * Unknown state
	 */
	private static final int STATUS_UNKNOWN = 0;
	
	/**
	 * Amarok status "pause"
	 */
	private static final int STATUS_PAUSE = 1;
	
	/**
	 * Amarok status "play"
	 */
	private static final int STATUS_PLAY = 2;
	
	/**
	 * The lines of text that will be displayed on screen
	 */
	private List<String> lines = new LinkedList<String>();
	
	/**
	 * default lines to display
	 */
	private static final String[] DEFAULT_LINES = {
		"%currentTimePad% %title%", //$NON-NLS-1$
		"%totalTimePad% %artist%", //$NON-NLS-1$
		"" //$NON-NLS-1$
	};
	
	/**
	 * Constant for the current time key
	 */
	private static final String CURRENT_TIME = "currentTime"; //$NON-NLS-1$
	
	/**
	 * Constant for the total time key
	 */
	private static final String TOTAL_TIME = "totalTime"; //$NON-NLS-1$
	
	/**
	 * Active tags, present in the configuration in the "%key%" form
	 */
	private Set<String> displayKeys = new LinkedHashSet<String>();
	
	/**
	 * The functions that we will call Amarok for, in order to obtain the values for the above keys
	 */
	private Set<String> amarokKeys = new LinkedHashSet<String>();
	
	/**
	 * Font to use
	 */
	private Font font;
	
	/**
	 * Font height, in pixels
	 */
	private int fontHeight = 0;
	
	/**
	 * Whether or not to draw the progress bar
	 */
	private boolean drawProgress = true;
	
	/**
	 * When scrolling, start with a pause
	 */
	private int[] iTextOffsets = new int[DEFAULT_LINES.length];
	
	/**
	 * How many iterations to pause at the beginning
	 */
	private int iTextPause = 3;
	
	/**
	 * How many characters to advance?
	 */
	private int iTextAdvance = 1;
	
	/**
	 * Scrollable keys
	 */
	private static final Set<String> SCROLLABLE_KEYS = new HashSet<String>();
	
	static{
		SCROLLABLE_KEYS.add("artist"); //$NON-NLS-1$
		SCROLLABLE_KEYS.add("comment"); //$NON-NLS-1$
		SCROLLABLE_KEYS.add("encodedURL"); //$NON-NLS-1$
		SCROLLABLE_KEYS.add("lyrics"); //$NON-NLS-1$
		SCROLLABLE_KEYS.add("lastfmStation"); //$NON-NLS-1$
		SCROLLABLE_KEYS.add("comment"); //$NON-NLS-1$
		SCROLLABLE_KEYS.add("path"); //$NON-NLS-1$
		SCROLLABLE_KEYS.add("title"); //$NON-NLS-1$
	}
	
	/**
	 * Get the real function to call from Amarok to obtain the value for a given configuration key
	 * 
	 * @param sConfigurationKey configuration tag
	 * @return the function to call
	 */
	private static final String getAmarokKey(final String sConfigurationKey){
		if (sConfigurationKey==null || sConfigurationKey.length()==0)
			return null;
		
		if (sConfigurationKey.equals("currentTimePad")) //$NON-NLS-1$
			return CURRENT_TIME;
		
		if (sConfigurationKey.equals("totalTimePad")) //$NON-NLS-1$
			return TOTAL_TIME;
		
		return sConfigurationKey;
	}
	
	/**
	 * Reset text position when reaching end of line or a new song is playing
	 * @param i index to reset 
	 */
	private void resetTextOffset(final int i){
		this.iTextOffsets[i] = -this.iTextAdvance * this.iTextPause;
	}
	
	@Override
	public void init(final ExtProperties config, final int xstart, final int ystart) {
		super.init(config, xstart, ystart);
		
		// start with the status. getAmarokStatus() will break after status if the output is not 0, 1 or 2
		this.amarokKeys.add("status"); //$NON-NLS-1$
		
		this.width  = config.geti("leds.modules.Amarok.width", 170); //$NON-NLS-1$
		this.height = config.geti("leds.modules.Amarok.height", 21); //$NON-NLS-1$
		
		this.font = new Font("Courier New", Font.PLAIN, config.geti("leds.modules.Amarok.fontsize", 10)); //$NON-NLS-1$ //$NON-NLS-2$
		
		this.drawProgress = config.getb("leds.modules.Amarok.progressBar", true); //$NON-NLS-1$
		
		this.iTextAdvance = config.geti("leds.modules.Amarok.scroll_advance", 1); //$NON-NLS-1$
		
		this.iTextPause = config.geti("leds.modules.Amarok.scroll_pause", 3); //$NON-NLS-1$
		
		for (int i=0; i<DEFAULT_LINES.length; i++)
			resetTextOffset(i);
		
		for (int i=1; i<=DEFAULT_LINES.length; i++){
			final String sLine = config.gets("leds.modules.Amarok.line"+i, DEFAULT_LINES[i-1]); //$NON-NLS-1$
			
			if (sLine.length()>0){
				this.lines.add(sLine);
				
				int idx = 0;
				
				while ( (idx=sLine.indexOf('%', idx))>=0 ){
					final int idx2 = sLine.indexOf('%', idx+1);
					
					if (idx2>0){
						final String sKey = sLine.substring(idx+1, idx2);
						
						this.displayKeys.add(sKey);
						
						final String sAmarokKey = getAmarokKey(sKey);
						if (sAmarokKey!=null)
							this.amarokKeys.add(sAmarokKey);
					}
					else
						break;
					
					idx = idx2+1;
				}
			}
			else
				break;
		}
		
		if (this.drawProgress){
			this.amarokKeys.add(CURRENT_TIME);
			this.amarokKeys.add(TOTAL_TIME);
		}
	}
	
	/**
	 * Previous values of each key, to know when to scroll
	 */
	private final HashMap<String, String> previous = new HashMap<String, String>();
	
	/**
	 * Fill a template line with a set of values, for a few of them take into account scrolling
	 * 
	 * @param sTemplate template line
	 * @param values values
	 * @param iTextOffset offset to scroll to
	 * @return the formatted line
	 */
	private String fillTemplate(final String sTemplate, final HashMap<String, String> values, final int iTextOffset){
		String ret = sTemplate;
		
		for (String sKey: this.displayKeys){
			String sValue = values.get(sKey);
			
			String sOldValue = this.previous.get(sKey);
			
			if (iTextOffset>0 && sValue!=null && SCROLLABLE_KEYS.contains(sKey) && sValue.equals(sOldValue)){
				sValue = iTextOffset < sValue.length() ? sValue.substring(iTextOffset)+" " : ""; //$NON-NLS-1$ //$NON-NLS-2$
			}
			
			ret = Utils.replace(ret, '%'+sKey+'%', sValue!=null ? sValue : ""); //$NON-NLS-1$
		}
		
		return ret;
	}
	
	@Override
	public boolean draw(final Canvas c) {
		final HashMap<String, String> status = getAmarokStatus();
		
		if (status==null || status.size()==0)
			return false;
		
		final Graphics2D graphics = c.getGraphics();
		
		final Font fOld = c.getGraphics().getFont();
		
		try{
			// draw only on the specified region, we might have weird strings to display
			graphics.setClip(this.x, this.y, this.width, this.height);
			
			// use the internal font
			c.getGraphics().setFont(this.font);
			
			if (this.fontHeight==0)
				this.fontHeight = c.getTextHeight();
			
			final int iStatus = Integer.parseInt(status.get("status")); //$NON-NLS-1$
			
			for (int i=0; i<this.lines.size(); i++){
				String sLine = fillTemplate(this.lines.get(i), status, 0);
				
				boolean tooLong = c.getTextWidth(sLine) > this.width;
				
				if (tooLong){
					this.iTextOffsets[i] += this.iTextAdvance;
					
					sLine = fillTemplate(this.lines.get(i), status, this.iTextOffsets[i]);
					
					if (c.getTextWidth(sLine) < this.width)
						resetTextOffset(i);
				}
				else{
					resetTextOffset(i);
				}
				
				graphics.drawString(sLine, this.x, this.y - 2 + (i+1)*(this.fontHeight-1));
			}
			
			for (String sKey : this.displayKeys){
				final String sValue = status.get(sKey);
				
				this.previous.put(sKey, sValue);
			}
			
			final String sTotalTime = status.get(TOTAL_TIME);
			final String sCurrentTime = status.get(CURRENT_TIME);
			
			if (this.drawProgress && sTotalTime!=null && sCurrentTime!=null){
				final int iLineXStart = this.x;
				final int iLineXEnd = this.x + this.width - 1;
				final int iLineY = this.y + this.height - 3;
				
				graphics.drawLine(iLineXStart, iLineY, iLineXEnd, iLineY);
				graphics.drawLine(iLineXStart, iLineY-2, this.x, iLineY+2);
				graphics.drawLine(iLineXEnd, iLineY-2, iLineXEnd, iLineY+2);
				
				int iTotalTime = decodeTime(sTotalTime);
				int iCurrentTime = decodeTime(sCurrentTime);
				
				if (iTotalTime < iCurrentTime)
					iTotalTime = iCurrentTime;
				
				if (iTotalTime > 0){
					final double percentage = (double) iCurrentTime / iTotalTime;
					
					final int where = iLineXStart + 2 + (int) Math.round(percentage * (this.width-4));
					
					switch (iStatus){						
						case STATUS_PLAY:
							graphics.drawLine(where-2, iLineY-2, where-2, iLineY+2);
							graphics.drawLine(where-1, iLineY-2, where-1, iLineY+2);
							graphics.drawLine(where, iLineY-1, where, iLineY+1);
							graphics.drawLine(where+1, iLineY-1, where+1, iLineY+1);
							break;
							
						case STATUS_PAUSE:
							graphics.drawRect(where-2, iLineY-2, 1, 5);
							graphics.setColor(Color.BLACK);
							graphics.drawLine(where, iLineY, where, iLineY);
							graphics.setColor(Color.WHITE);
							graphics.drawRect(where+1, iLineY-2, 1, 5);
							break;
							
						default:
							graphics.drawLine(where, iLineY-2, where, iLineY+2);								
					}
				}
			}
		}
		finally{
			// restore the old font
			graphics.setFont(fOld);
			
			// re-enable full area drawing, for other modules  
			graphics.setClip(null);
		}
		
		return true;
	}
	
	/**
	 * Convert a time representation like (HH:)MM:ss into seconds
	 * 
	 * @param sTime time description in Amarok style 
	 * @return seconds, or negative number if decoding was not possible
	 */
	@SuppressWarnings("fallthrough")
	private static final int decodeTime(final String sTime){
		if (sTime==null || sTime.length()==0)
			return -1;
		
		// As far as I can tell Amarok displays time in (HH:)MM:ss, but maybe in the future they will add the days in front...
		final StringTokenizer st = new StringTokenizer(sTime, ": d"); //$NON-NLS-1$
		
		int iDays    = 0;
		int iSeconds = 0;
		int iMinutes = 0;
		int iHours   = 0;
		
		try{
			switch (st.countTokens()){
				case 4:
					iDays    = Integer.parseInt(st.nextToken());
				case 3:
					iHours   = Integer.parseInt(st.nextToken());
				case 2:
					iMinutes = Integer.parseInt(st.nextToken());
				case 1:
					iSeconds = Integer.parseInt(st.nextToken());
			}
			
			return ((iDays * 24 + iHours) * 60 + iMinutes) * 60 + iSeconds;
		}
		catch (Exception e){
			// ignore
		}
		
		return -1;
	}
	
	/**
	 * What to query first? DCOP or DBUS ?
	 */
	private boolean bDCOPFirst = true;
	
	/**
	 * @return value for each key in {@link #displayKeys}
	 */
	private HashMap<String, String> getAmarokStatus() {
		HashMap<String, String> ret = null;
		
		if (this.bDCOPFirst){
			ret = getAmarokStatusDCOP();
			
			if (ret==null){
				ret = getAmarokStatusDBUS();
				
				if (ret!=null)
					this.bDCOPFirst = false;
			}
		}
		else{
			ret = getAmarokStatusDBUS();
			
			if (ret==null){
				ret = getAmarokStatusDCOP();
				
				if (ret!=null)
					this.bDCOPFirst = true;
			}
		}
		
		if (ret!=null)
			postProcess(ret);
		
		return ret;
	}

	/**
	 * DBUS interface query
	 * 
	 * @return all status values
	 */
	private HashMap<String, String> getAmarokStatusDBUS(){
		final DBUSApp amarok = DBUSApp.getApplication("org.kde.amarok"); //$NON-NLS-1$
		
		if (amarok==null)
			return null;
		
		final HashMap<String, String> ret = new HashMap<String, String>();
		
		String sMetadata = amarok.execute("/Player", "org.freedesktop.MediaPlayer.GetMetadata"); //$NON-NLS-1$ //$NON-NLS-2$
		
		if (sMetadata.length()==0)
			return ret;
		
		final BufferedReader br = new BufferedReader(new StringReader(sMetadata));
		
		String sLine;
		
		try {
			while ( (sLine=br.readLine()) != null ){
				final int idx = sLine.indexOf(':');
				
				if (idx>0){
					String sKey = sLine.substring(0, idx).trim();
					String sValue = sLine.substring(idx+1).trim();
					
					if (sKey.equals("mtime")){ //$NON-NLS-1$
						sKey = TOTAL_TIME;
						sValue = Kaffeine.toInterval(Integer.parseInt(sValue)/1000, false);
					}
						
					ret.put(sKey, sValue);
				}
			}
		}
		catch (IOException e) {
			// ignore
		}
		
		try {
			final String sValue = Kaffeine.toInterval(Integer.parseInt(amarok.execute("/Player", "org.freedesktop.MediaPlayer.PositionGet"))/1000, false);  //$NON-NLS-1$//$NON-NLS-2$
			
			ret.put(CURRENT_TIME, sValue);
		}
		catch (Exception e){
			// ignore
		}
		
		ret.put("status", ""+STATUS_UNKNOWN); //$NON-NLS-1$ //$NON-NLS-2$
		
		return ret;
	}
	
	/**
	 * DCOP interface query
	 * 
	 * @return value for each key in {@link #displayKeys}
	 */
	private HashMap<String, String> getAmarokStatusDCOP() {
		final DCOPApp amarok = Utils.getDCOPApp("amarok"); //$NON-NLS-1$
		
		if (amarok==null)
			return null;

		final HashMap<String, String> values = new HashMap<String, String>();
		
		for (String sKey: this.amarokKeys){
			String sOutput = amarok.execute("player", sKey);  //$NON-NLS-1$
			
			if (sOutput!=null && sOutput.length()>0){
				if (sOutput.indexOf('\n')>=0)
					sOutput = sOutput.substring(0, sOutput.indexOf('\n'));
				
				if (sKey.indexOf("Time")>0){ //$NON-NLS-1$
					if (decodeTime(sOutput)<0){
						// the time has some incorrect format (usually "?"), so we don't consider it
						continue;
					}
				}
				
				if (sOutput.length()>0)
					values.put(sKey, sOutput);
			}
			
			if (sKey.equals("status")){ //$NON-NLS-1$
				try{
					final int iStatus = Integer.parseInt(sOutput);
					
					if (iStatus!=STATUS_PLAY && iStatus!=STATUS_PAUSE)
						return null;
				}
				catch (Throwable t){
					// cannot talk to Amarok over dcop
					return null;
				}
			}
		}
		
		return values;
	}
	
	/**
	 * Create some useful derived values
	 * 
	 * @param values values
	 */
	private static void postProcess(final HashMap<String, String> values){
		if (!values.containsKey("artist")){ //$NON-NLS-1$
			String sTitle = values.get("title"); //$NON-NLS-1$
			
			int idx;
			
			if (sTitle!=null && (idx=sTitle.indexOf(" - "))>0){ //$NON-NLS-1$
				final String sArtist = sTitle.substring(0, idx).trim();
				sTitle = sTitle.substring(idx+3).trim();
				
				values.put("artist", sArtist); //$NON-NLS-1$
				values.put("title", sTitle); //$NON-NLS-1$
			}
		}
		
		String sTotalTime = values.get(TOTAL_TIME);
		String sCurrentTime = values.get(CURRENT_TIME);
		
		String sTotalTimePad = sTotalTime;
		
		if (sTotalTimePad==null)
			sTotalTimePad = ""; //$NON-NLS-1$
			
		if (sCurrentTime!=null){
			while (sTotalTimePad.length() < sCurrentTime.length())
				sTotalTimePad = ' '+sTotalTimePad;
		}
		
		// totalTimePad will have at least the same length as currentTime
		values.put("totalTimePad", sTotalTimePad); //$NON-NLS-1$
		
		if (sTotalTime!=null && sCurrentTime==null){
			sCurrentTime = "0:00"; //$NON-NLS-1$
			values.put(CURRENT_TIME, sCurrentTime);
		}
		
		String sCurrentTimePad = sCurrentTime;
		
		if (sCurrentTimePad==null)
			sCurrentTimePad = ""; //$NON-NLS-1$
		
		if (sTotalTime!=null){
			while (sCurrentTimePad.length() < sTotalTime.length())
				sCurrentTimePad = ' '+sCurrentTimePad;
		}
		
		// currentTimePad will have at least the same length as totalTime
		// this means that now currentTimePad.length() == totalTimePad.length()
		values.put("currentTimePad", sCurrentTimePad); //$NON-NLS-1$
	}
}
