/**
 * What's Kaffeine playing now ? 
 */
package leds.modules;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.util.LinkedList;
import java.util.List;

import leds.Canvas;
import leds.DCOPApp;
import leds.ExtProperties;
import leds.Utils;

/**
 * This module displays what Kaffeine is playing now
 * 
 * @author costing
 * @since Dec 23, 2008
 */
public class Kaffeine extends BaseModule {
	/**
	 * default lines to display
	 */
	private static final String[] DEFAULT_LINES = {
		"%title%", //$NON-NLS-1$
		"%currentTime% / %totalTime% (-%remainingTime%)", //$NON-NLS-1$
		"" //$NON-NLS-1$
	};
	
	/**
	 * Font to use
	 */
	private Font font;
	
	/**
	 * Font height, in pixels
	 */
	private int fontHeight = 0;
	
	/**
	 * Whether or not to draw the progress bar
	 */
	private boolean drawProgress = true;
	
	/**
	 * The lines of text that will be displayed on screen
	 */
	private List<String> lines = new LinkedList<String>();
	
	
	/* (non-Javadoc)
	 * @see leds.modules.BaseModule#init(leds.ExtProperties, int, int)
	 */
	@Override
	public void init(ExtProperties config, int xStart, int yStart) {
		super.init(config, xStart, yStart);
		
		this.width  = config.geti("leds.modules.Kaffeine.width", 170); //$NON-NLS-1$
		this.height = config.geti("leds.modules.Kaffeine.height", 21); //$NON-NLS-1$
		
		this.font = new Font("Courier New", Font.PLAIN, config.geti("leds.modules.Kaffeine.fontsize", 10)); //$NON-NLS-1$ //$NON-NLS-2$
		
		this.drawProgress = config.getb("leds.modules.Kaffeine.progressBar", true); //$NON-NLS-1$
		
		for (int i=1; i<=DEFAULT_LINES.length; i++){
			final String sLine = config.gets("leds.modules.Kaffeine.line"+i, DEFAULT_LINES[i-1]); //$NON-NLS-1$
			
			if (sLine.length()>0){
				this.lines.add(sLine);
			}
		}
	}
	
	/**
	 * Old position in the stream, in seconds. Based on this and the current position we will draw either a PLAY or a PAUSE on the progress bar.
	 */
	private int iOldPos = -1;
	
	/* (non-Javadoc)
	 * @see leds.Module#draw(leds.Canvas)
	 */
	@Override
	public boolean draw(final Canvas c) {
		if (getKaffeineBool("isPlaying", false) == false){ //$NON-NLS-1$
			// Kaffeine is not running / playing anything now
			this.iOldPos = -1;
			
			return false;
		}
		
		final String sTitle = getKaffeineDCOPValue("title"); //$NON-NLS-1$
		
		final Graphics2D graphics = c.getGraphics();
		
		final Font fOld = c.getGraphics().getFont();
		
		final int iTotalTime = getKaffeineInt("getLength"); //$NON-NLS-1$
		
		final int iCurrentPos = getKaffeineInt("getTimePos"); //$NON-NLS-1$
		
		try{
			// draw only on the specified region, we might have weird strings to display
			graphics.setClip(this.x, this.y, this.width, this.height);
			
			// use the internal font
			c.getGraphics().setFont(this.font);
			
			if (this.fontHeight==0)
				this.fontHeight = c.getTextHeight();
			
			final boolean bForceHours = iTotalTime >= 3600 || iCurrentPos >= 3600;
			
			final String sTotalTime = toInterval(iTotalTime, bForceHours);
			final String sCurrentTime = toInterval(iCurrentPos, bForceHours);
			final String sRemainingTime = toInterval(iTotalTime - iCurrentPos, bForceHours);
			
			for (int i=0; i<this.lines.size(); i++){
				String sLine = this.lines.get(i);
				
				sLine = Utils.replace(sLine, "%title%", sTitle); //$NON-NLS-1$
				sLine = Utils.replace(sLine, "%currentTime%", sCurrentTime); //$NON-NLS-1$
				sLine = Utils.replace(sLine, "%totalTime%", sTotalTime); //$NON-NLS-1$
				sLine = Utils.replace(sLine, "%remainingTime%", sRemainingTime); //$NON-NLS-1$
				
				// first line: the title
				graphics.drawString(sLine, this.x, this.y - 2 + (i+1)*(this.fontHeight-1));
			}
			
			if (this.drawProgress && iTotalTime >= 0 && iCurrentPos >= 0){
				final double percentage = (double) iCurrentPos / iTotalTime;
				
				final int iLineXStart = this.x;
				final int iLineXEnd = this.x + this.width - 1;
				final int iLineY = this.y + this.height - 3;
				
				graphics.drawLine(iLineXStart, iLineY, iLineXEnd, iLineY);
				graphics.drawLine(iLineXStart, iLineY-2, this.x, iLineY+2);
				graphics.drawLine(iLineXEnd, iLineY-2, iLineXEnd, iLineY+2);
				
				final int where = iLineXStart + 2 + (int) Math.round(percentage * (this.width-4));
				
				if (iCurrentPos != this.iOldPos){
					// we have advanced, we are playing something
					
					graphics.drawLine(where-2, iLineY-2, where-2, iLineY+2);
					graphics.drawLine(where-1, iLineY-2, where-1, iLineY+2);
					graphics.drawLine(where, iLineY-1, where, iLineY+1);
					graphics.drawLine(where+1, iLineY-1, where+1, iLineY+1);
					
					this.iOldPos = iCurrentPos;
				}
				else{
					// it's paused
					
					graphics.drawRect(where-2, iLineY-2, 1, 5);
					graphics.setColor(Color.BLACK);
					graphics.drawLine(where, iLineY, where, iLineY);
					graphics.setColor(Color.WHITE);
					graphics.drawRect(where+1, iLineY-2, 1, 5);
				}
			}
		}
		finally{
			// restore the old font
			graphics.setFont(fOld);
			
			// re-enable full area drawing, for other modules  
			graphics.setClip(null);
		}
			
		return true;
	}
	
	/**
	 * @param iInterval interval length, in seconds
	 * @param bForceHours force displaying of hours, even if it's zero
	 * @return human representation
	 */
	static String toInterval(final int iInterval, final boolean bForceHours) {
		if (iInterval < 0)
			return ""; //$NON-NLS-1$
		
		final int iSeconds = iInterval % 60;
		final int iMinutes = (iInterval / 60) % 60;
		final int iHours = iInterval / 3600;
		
		return (iHours>0 || bForceHours ? iHours + ":" : "" ) + showZero(iMinutes) + ':' + showZero(iSeconds);   //$NON-NLS-1$ //$NON-NLS-2$
	}
	
	/**
	 * Add a 0 in front, if required
	 * 
	 * @param value value to display
	 * @return string
	 */
	private static String showZero(final int value){
		return (value<10 ? "0" : "")+value; //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Get the value of a Kaffeine int function
	 * 
	 * @param sFunction function to call
	 * @return parsed value, or -1 if the text cannot be converted to int
	 */
	private int getKaffeineInt(final String sFunction){
		try{
			return Integer.parseInt(getKaffeineDCOPValue(sFunction));
		}
		catch (Exception e){
			// ignore
		}
		
		return -1;
	}

	/**
	 * Get the value of a Kaffeine boolean function
	 * 
	 * @param sFunction function to call
	 * @param bDefault default value to return if the value cannot be parsed
	 * @return parsed value, or default
	 */
	private boolean getKaffeineBool(final String sFunction, final boolean bDefault){
		return Utils.stringToBool(getKaffeineDCOPValue(sFunction), bDefault); 
	}
	
	/**
	 * Call the DCOP function and return the output
	 * 
	 * @param sFunction function to call from Kaffeine interface
	 * @return value (never <code>null</code>)
	 */
	private String getKaffeineDCOPValue(final String sFunction){
		final DCOPApp kaffeine = Utils.getDCOPApp("kaffeine"); //$NON-NLS-1$
		
		if (kaffeine==null)
			return ""; //$NON-NLS-1$
		
		String sOutput = kaffeine.execute("KaffeineIface", sFunction); //$NON-NLS-1$
		
		if (sOutput!=null){
			if (sOutput.indexOf('\n')>=0)
				sOutput = sOutput.substring(0, sOutput.indexOf('\n'));
		}
		else
			sOutput = ""; //$NON-NLS-1$
		
		return sOutput;
	}
	
}
