/**
 * 
 */
package leds.modules;

import java.awt.Graphics2D;
import java.text.DecimalFormat;

import leds.Canvas;
import leds.ExtProperties;
import leds.Utils;
import leds.Values;

/**
 * Show the aggregated network traffic since the last draw, based on the information from /proc/net/dev.
 * 
 * @author costing
 * @since Oct 20, 2008
 * @see Utils#getNetIO()
 */
public class NetIO extends BaseModule {

	/**
	 * Whether or not to subtract the loopback interface details
	 */
	private boolean bSubstractLoopback = true;
	
	/**
	 * Number formatting
	 */
	private DecimalFormat format = null;
	
	/**
	 * Whether or not to show both IN and OUT with the same measurement units
	 */
	private boolean bSameSize = true;
	
	/**
	 * Initialize the counters 
	 */
	public NetIO() {
		Utils.getNetIO();
	}
	
	/* (non-Javadoc)
	 * @see leds.modules.BaseModule#init(leds.ExtProperties, int, int)
	 */
	@Override
	public void init(ExtProperties config, int xstart, int ystart) {
		super.init(config, xstart, ystart);
		
		this.bSubstractLoopback = config.getb("leds.modules.NetIO.substract_loopback", true); //$NON-NLS-1$
		
		this.format = new DecimalFormat(config.gets("leds.modules.NetIO.format", "###"));  //$NON-NLS-1$//$NON-NLS-2$
		
		this.bSameSize = config.getb("leds.modules.NetIO.same_size", true); //$NON-NLS-1$
	}
	
	/**
	 * Convert a network speed (in bytes/second) in a nice human-readable format
	 * 
	 * @param speed speed, in bytes/second
	 * @param digits number of digits (the result of Math.log10(speed))
	 * @return human readable format
	 */
	private String speedToString(final double speed, final double digits){
		if( digits >= 3 && digits < 6  )
			return this.format.format( speed / 1024 ) + " KB"; //$NON-NLS-1$
		
		if( digits >= 6 && digits < 9  )
			return this.format.format( speed / (1024*1024) ) + " MB"; //$NON-NLS-1$
		
		if( digits >= 9 && digits < 12  )
			return this.format.format( speed / (1024*1024*1024) ) + " GB"; //$NON-NLS-1$
		
		return this.format.format( speed ) + " B"; //$NON-NLS-1$
	}
	
	/* (non-Javadoc)
	 * @see leds.Module#draw(leds.Canvas)
	 */
	@Override
	public boolean draw(final Canvas c) {
		if (this.width == 0){
			final double ref = 123456789;
			
			this.width = c.getTextWidth(speedToString(ref, Math.log10(ref))+" ") + 8; //$NON-NLS-1$
		}
		
		if (this.height == 0)
			this.height = 2*(c.getTextHeight() - 2) + 2;
		
		final Values io = Utils.getNetIO();
		
		final int where = this.x + this.width;
		final int Y1 = this.y + c.getTextHeight() - 2;
		final int Y2 = this.y + this.height; 
		
		double dTotalOut = io.getz("TOTAL_OUT"); //$NON-NLS-1$
		double dTotalIn = io.getz("TOTAL_IN"); //$NON-NLS-1$
		
		if (this.bSubstractLoopback){
			dTotalOut -= io.getz("lo_OUT"); //$NON-NLS-1$
			dTotalIn -= io.getz("lo_IN"); //$NON-NLS-1$
		}
		
		final double digitsOut = Math.log10(dTotalOut);
		final double digitsIn = Math.log10(dTotalIn);
		
		c.drawString(speedToString(dTotalOut, this.bSameSize ? Math.max(digitsIn, digitsOut) : digitsOut), where-8, Y1, Canvas.ALIGN_RIGHT);
		c.drawString(speedToString(dTotalIn, this.bSameSize ? Math.max(digitsIn, digitsOut) : digitsIn), where-8, Y2, Canvas.ALIGN_RIGHT);

		final Graphics2D graphics = c.getGraphics();

		// arrows up and down
		graphics.drawLine(where-3, Y1 - 8, where-3, Y1 - 1);
		graphics.drawLine(where-6, Y1 - 5, where-3, Y1 - 8);
		graphics.drawLine(where  , Y1 - 5, where-3, Y1 - 8);

		graphics.drawLine(where-3, Y2 - 8, where-3, Y2 - 1);
		graphics.drawLine(where-6, Y2 - 4, where-3, Y2 - 1);
		graphics.drawLine(where  , Y2 - 4, where-3, Y2 - 1);
		
		return true;
	}

}
